using System;

// ReSharper disable MemberCanBeMadeStatic.Local

namespace IdeaBlade.EntityModel {

  public static class EntitySaveOperationExtensions {

    public static Func<EntitySaveOperation, IEntitySaveOperationWrapper> OperationWrapperFactory =
      op => new DefaultEntitySaveOperationWrapper(op);

    private static IEntitySaveOperationWrapper MakeWrapper(EntitySaveOperation op) {
      return OperationWrapperFactory(op);
    }

    public static IEntitySaveOperationWrapper OnCancelled(this EntitySaveOperation op, Action action) {
      return MakeWrapper(op).OnCancelled(action);
    }

    /// <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
    public static void OnCompleted(this EntitySaveOperation op, Action<EntitySavedEventArgs> action) {
      MakeWrapper(op).OnCompleted(action);
    }

    public static IEntitySaveOperationWrapper OnError(this EntitySaveOperation op, Action<Exception> action) {
      return MakeWrapper(op).OnError(action);
    }

    public static IEntitySaveOperationWrapper OnError(this EntitySaveOperation op, string failureMessageTemplate = null) {
      return MakeWrapper(op).OnError(failureMessageTemplate);
    }

    public static IEntitySaveOperationWrapper OnSuccess(this EntitySaveOperation op, Action<EntitySavedEventArgs> action) {
      return MakeWrapper(op).OnSuccess(action);
    }
  }

  public interface IEntitySaveOperationWrapper {

    IEntitySaveOperationWrapper OnSuccess(Action<EntitySavedEventArgs> action);
    IEntitySaveOperationWrapper OnCancelled(Action action);
    /// <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
    void OnCompleted(Action<EntitySavedEventArgs> action);
    IEntitySaveOperationWrapper OnError(Action<Exception> action);
    IEntitySaveOperationWrapper OnError(string failureMessageTemplate = null);

  }

  public class DefaultEntitySaveOperationWrapper : IEntitySaveOperationWrapper {

    public DefaultEntitySaveOperationWrapper(EntitySaveOperation op) {
      Operation = op;
      OnCompletedHandler = new OperationOnCompletedHandler<EntitySaveOperation, EntitySavedEventArgs>(op);
      op.Completed += OnCompletedHandler.CompletedHandler;
    }

    public EntitySaveOperation Operation { get; private set; }

    protected OperationOnCompletedHandler<EntitySaveOperation, EntitySavedEventArgs> OnCompletedHandler { get; set; }

    public IEntitySaveOperationWrapper OnCancelled(Action action) {
      OnCompletedHandler.OnCancelled(action);
      return this;
    }

    public void OnCompleted(Action<EntitySavedEventArgs> action) {
      OnCompletedHandler.OnCompleted(action);
    }

    public IEntitySaveOperationWrapper OnError(Action<Exception> action) {
      OnCompletedHandler.OnError(action);
      return this;
    }

    public IEntitySaveOperationWrapper OnError(string failureMessageTemplate = null) {
      OnCompletedHandler.OnError(MakeFailureMessageTemplate(failureMessageTemplate));
      return this;
    }

    private string MakeFailureMessageTemplate(string failureMessageTemplate) {
      if (String.IsNullOrWhiteSpace(failureMessageTemplate)) {
        failureMessageTemplate = "Save failed with exception \"{0}\".";
      }
      return failureMessageTemplate;
    }

    public IEntitySaveOperationWrapper OnSuccess(Action<EntitySavedEventArgs> action) {
      OnCompletedHandler.OnSuccess(action);
      return this;
    }

  }

}