using System;

// ReSharper disable MemberCanBeMadeStatic.Local

namespace IdeaBlade.EntityModel {

  public static class InvokeServerMethodOperationExtensions {

    public static Func<InvokeServerMethodOperation, IInvokeServerMethodOperationWrapper> OperationWrapperFactory =
      op => new DefaultInvokeServerMethodOperationWrapper(op);


    private static IInvokeServerMethodOperationWrapper MakeWrapper(InvokeServerMethodOperation op) {
      return OperationWrapperFactory(op);
    }

    public static IInvokeServerMethodOperationWrapper OnCancelled(this InvokeServerMethodOperation op, Action action) {
      return MakeWrapper(op).OnCancelled(action);
    }

    /// <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
    public static void OnCompleted(this InvokeServerMethodOperation op, Action<InvokeServerMethodEventArgs> action) {
      MakeWrapper(op).OnCompleted(action);
    }

    public static IInvokeServerMethodOperationWrapper OnError(this InvokeServerMethodOperation op, Action<Exception> action) {
      return MakeWrapper(op).OnError(action);
    }

    public static IInvokeServerMethodOperationWrapper OnError(this InvokeServerMethodOperation op, string failureMessageTemplate = null) {
      return MakeWrapper(op).OnError(failureMessageTemplate);
    }

    public static IInvokeServerMethodOperationWrapper OnSuccess(this InvokeServerMethodOperation op, Action<InvokeServerMethodEventArgs> action) {
      return MakeWrapper(op).OnSuccess(action);
    }
  }

  public interface IInvokeServerMethodOperationWrapper {

    IInvokeServerMethodOperationWrapper OnSuccess(Action<InvokeServerMethodEventArgs> action);
    IInvokeServerMethodOperationWrapper OnCancelled(Action action);
    /// <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
    void OnCompleted(Action<InvokeServerMethodEventArgs> action);
    IInvokeServerMethodOperationWrapper OnError(Action<Exception> action);
    IInvokeServerMethodOperationWrapper OnError(string failureMessageTemplate = null);

  }

  public class DefaultInvokeServerMethodOperationWrapper : IInvokeServerMethodOperationWrapper {

    public DefaultInvokeServerMethodOperationWrapper(InvokeServerMethodOperation op) {
      Operation = op;
      OnCompletedHandler = new OperationOnCompletedHandler<InvokeServerMethodOperation, InvokeServerMethodEventArgs>(op);
      op.Completed += OnCompletedHandler.CompletedHandler;
    }

    public InvokeServerMethodOperation Operation { get; private set; }

    protected OperationOnCompletedHandler<InvokeServerMethodOperation, InvokeServerMethodEventArgs> OnCompletedHandler { get; set; }

    public IInvokeServerMethodOperationWrapper OnCancelled(Action action) {
      OnCompletedHandler.OnCancelled(action);
      return this;
    }

    public void OnCompleted(Action<InvokeServerMethodEventArgs> action) {
      OnCompletedHandler.OnCompleted(action);
    }

    public IInvokeServerMethodOperationWrapper OnError(Action<Exception> action) {
      OnCompletedHandler.OnError(action);
      return this;
    }

    public IInvokeServerMethodOperationWrapper OnError(string failureMessageTemplate = null) {
      OnCompletedHandler.OnError(MakeFailureMessageTemplate(failureMessageTemplate));
      return this;
    }

    private string MakeFailureMessageTemplate(string failureMessageTemplate) {
      if (String.IsNullOrWhiteSpace(failureMessageTemplate)) {
        failureMessageTemplate = "Server Method failed with exception \"{0}\".";
      }
      return failureMessageTemplate;
    }

    public IInvokeServerMethodOperationWrapper OnSuccess(Action<InvokeServerMethodEventArgs> action) {
      OnCompletedHandler.OnSuccess(action);
      return this;
    }

  }

}