﻿using System;

namespace IdeaBlade.EntityModel {
  public class OperationOnCompletedHandler<TOperation, TEventArgs>
    where TOperation : BaseOperation
    where TEventArgs : AsyncEventArgs {

    public OperationOnCompletedHandler(TOperation op) {
      Operation = op;
      ClearActions();
    }

    protected TOperation Operation { get; private set; }

    public virtual void CompletedHandler(object sender, TEventArgs args) {
      try {
        if (null == args) {
          throw new ArgumentNullException("args");
        }
        CompletedEventArgs = args;
        DoOnCancelled();
        DoOnError();
        DoOnSuccess();
        DoOnCompleted();
      } finally {
        ClearActions();
      }
    }

    #region DoOn ...

    protected virtual void DoOnCancelled(Action action = null) {
      if (CompletedEventArgs.Cancelled) {
        _onCancelledActions();
        if (null != action) action();
      }
    }

    protected virtual void DoOnError(Action<Exception> action = null) {
      if (CompletedEventArgs.HasError) {
        Operation.MarkErrorAsHandled();
        _onErrorActions(CompletedEventArgs.Error);
        if (null != action) action(CompletedEventArgs.Error);
      }
    }

    protected virtual void DoOnSuccess(Action<TEventArgs> action = null) {
      if (CompletedEventArgs.CompletedSuccessfully) {
        _onSuccessActions(CompletedEventArgs);
        if (null != action) action(CompletedEventArgs);
      }
    }

    protected virtual void DoOnCompleted(Action<TEventArgs> action = null) {
      _onCompletedActions(CompletedEventArgs);
      if (null != action) action(CompletedEventArgs);
    }

    #endregion

    #region On ...

    public virtual void OnCancelled(Action action) {
      if (WasCompletedHandlerCalled) {
        DoOnCancelled(action);
      } else {
        if (null == action) return;
        _onCancelledActions += action;
      }
    }

    public virtual void OnCompleted(Action<TEventArgs> action) {
      if (WasCompletedHandlerCalled) {
        DoOnCompleted(action);
      } else {
        if (null == action) return;
        _onCompletedActions += action;
      }
    }

    public virtual void OnError(string failureMessageTemplate) {
      var action = FailureMessageAction(failureMessageTemplate);
      OnError(action);
    }

    /// <summary>
    /// Function taking a string to format the operation exception message and returning 
    /// an action that will take an exception, format it, and do something.
    /// </summary>
    /// <remarks>
    /// The default implementation will format the operation exception message
    /// and throw an <see cref="EntityServerException"/> with this message and the
    /// incoming exception in its inner exception.
    /// <para>
    /// <emphasis>
    /// Beware: may change the type of this exception in future as it
    /// may not always true that the EntityServer is the source.</emphasis>
    /// </para><para>
    /// You can replace to give different behavior in your environment.
    /// </para>
    /// </remarks>
    public static Func<string, Action<Exception>> FailureMessageAction =
      fmt => ex => {
               throw new EntityServerException(
                 String.Format(fmt, ex.GetType().Name + " - " + ex.Message), 
                 ex);
             }; 

    public virtual void OnError(Action<Exception> action) {
      if (WasCompletedHandlerCalled) {
        DoOnError(action);
      } else {
        if (null == action) return;
        _onErrorActions += action;
      }
    }

    public virtual void OnSuccess(Action<TEventArgs> action) {
      if (WasCompletedHandlerCalled) {
        DoOnSuccess(action);
      } else {
        if (null == action) return;
        _onSuccessActions += action;
      }
    }

    #endregion


    protected bool WasCompletedHandlerCalled { get { return null != CompletedEventArgs; } }
    protected TEventArgs CompletedEventArgs { get; private set;  }

    private void ClearActions() {
      _onCancelledActions = delegate {};
      _onCompletedActions = delegate {};
      _onErrorActions= delegate {};
      _onSuccessActions = delegate {};
    }

    private Action _onCancelledActions;
    private Action<TEventArgs> _onCompletedActions;
    private Action<Exception> _onErrorActions;
    private Action<TEventArgs> _onSuccessActions;

    }
}