using System;
using System.Collections.Generic;
using System.Linq;
using IdeaBlade.EntityModel;

namespace IdeaBlade.TestFramework {
  public static class TestingExtensions {

    /// <summary>
    /// Asserts that this <see cref="IEnumerable{T}"/> does not contain any items. 
    /// </summary>
    public static IEnumerable<T> ShouldBeEmpty<T>(
      this IEnumerable<T> source, string message = null, params object[] parameters) {
      if (source.Any()) {
        AssertHelper.HandleFail("ShouldBeEmpty", message, parameters);
      }
      return source;
    }

    /// <summary>
    /// Asserts that this <see cref="IEnumerable{T}"/> contains items. 
    /// </summary>
    public static IEnumerable<T> ShouldNotBeEmpty<T>(
      this IEnumerable<T> source, string message = null, params object[] parameters) {
      if (!source.Any()) {
        AssertHelper.HandleFail("ShouldNotBeEmpty", message, parameters);
      }
      return source;
    }

    //----------------

    /// <summary>
    /// Asserts that this string contains a target string. 
    /// </summary>
    public static string ShouldContain(this string source, string target, string message = null, params object[] parameters) {
      if (!source.Contains(target)) {
        AssertHelper.HandleFail("ShouldContain", message, parameters);
      }
      return source;
    }

    /// <summary>
    /// Asserts that this string does not contains a target string. 
    /// </summary>
    public static string ShouldNotContain(this string source, string target, string message = null, params object[] parameters) {
      if (source.Contains(target)) {
        AssertHelper.HandleFail("ShouldNotContain", message, parameters);
      }
      return source;
    }

    //----------------

    /// <summary>
    /// Asserts that this <see cref="IEnumerable{T}"/> contains a specific item. 
    /// </summary>
    public static IEnumerable<T> ShouldContain<T>(this IEnumerable<T> source, T item, string message = null, params object[] parameters) {
      if (!source.Contains(item)) {
        AssertHelper.HandleFail("ShouldContain", message, parameters);
      }
      return source;
    }

    /// <summary>
    /// Asserts that this <see cref="IEnumerable{T}"/> does not contain a specific item. 
    /// </summary>
    public static IEnumerable<T> ShouldNotContain<T>(this IEnumerable<T> source, T item, string message = null, params object[] parameters) {
      if (source.Contains(item)) {
        AssertHelper.HandleFail("ShouldNotContain", message, parameters);
      }
      return source;
    }

    //----------------

    /// <summary>
    /// Asserts that this actual <see cref="IEnumerable{T}"/> 
    /// contains all of the items in the expected list. 
    /// </summary>
    public static IEnumerable<T> ShouldContainAllOf<T>(
      this IEnumerable<T> actual, IEnumerable<T> expected, string message = null, params object[] parameters) {
      if (expected.Any(exp => !actual.Contains(exp))) {
        AssertHelper.HandleFail("ShouldContainAll", message, parameters);
      }
      return expected;
    }

    /// <summary>
    /// Asserts that this actual <see cref="IEnumerable{T}"/> 
    /// does not contain any of the items in the expected list. 
    /// </summary>
    public static IEnumerable<T> ShouldNotContainAnyOf<T>(
      this IEnumerable<T> actual, IEnumerable<T> expected, string message = null, params object[] parameters) {
      if (expected.Any(exp => actual.Contains(exp))) {
        AssertHelper.HandleFail("ShouldNotContainAnyOf", message, parameters);
      }
      return expected;
    }

    //----------------

    /// <summary>
    /// Asserts that this entity is either null or the null entity. 
    /// </summary>
    public static T ShouldBeNullOrNullEntity<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {
      return actual.AssertOnNullOrNullEntity(true, message, parameters);
    }

    /// <summary>
    /// Asserts that this entity is neither null nor the null entity. 
    /// </summary>
    public static T ShouldNotBeNullOrNullEntity<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {
      return actual.AssertOnNullOrNullEntity(false, message, parameters);
    }

    private static T AssertOnNullOrNullEntity<T>(
      this T actual, bool shouldBe = true, string message = null, params object[] parameters)
      where T : IEntity {

      var isNull = null == actual || actual.EntityAspect.IsNullEntity;
      if (shouldBe == isNull) return actual;

      const string stateName = "null or null entity";
      var shouldText = (shouldBe ? "Should be " : "Should not be ") + stateName;

      if (0 == parameters.Length) {
        message = String.Format("{0} of type '{1}'{2}.", 
          shouldText, typeof(T).Name, message);
      } 
      var assertionName = shouldText.Replace(" ", String.Empty);
      AssertHelper.HandleFail(assertionName, message, parameters);

      return actual;
    }

    //========= ENTITYSTATE ASSERTIONS ==========

    private static T AssertOnEntityStateTest<T>(
      this T actual, string stateName, Func<bool> predicate, bool shouldBe = true, 
      string message = null, params object[] parameters) 
      where T : IEntity {

      actual.ShouldNotBeNullOrNullEntity();
      if (shouldBe == predicate()) return actual;

      var shouldText = (shouldBe ? "Should be " : "Should not be ") + stateName;

      if (0 == parameters.Length) {
        message = String.Format(
          "{0} but EntityState is '{1}'{2}.",
          shouldText, actual.EntityAspect.EntityState, message);  
      } 
      var assertionName = shouldText.Replace(" ", String.Empty);
      AssertHelper.HandleFail(assertionName, message, parameters);

      return actual;
    }

    /// <summary>
    /// Asserts that this entity is Changed. 
    /// </summary>
    public static T ShouldBeChanged<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {

      return actual.AssertOnEntityStateTest("Changed", 
        ()=> actual.EntityAspect.IsChanged, true, 
        message, parameters);
    }

    /// <summary>
    /// Asserts that this entity is not Changed. 
    /// </summary>
    public static T ShouldNotBeChanged<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {

      return actual.AssertOnEntityStateTest("Changed", 
        () => actual.EntityAspect.IsChanged, false,
        message, parameters);
    }

    /// <summary>
    /// Asserts that this entity is Unchanged. 
    /// </summary>
    /// <remarks>
    /// Should be equivalent to <see cref="ShouldNotBeChanged{T}(T, string , object[])"/>
    /// although implemented in terms of the <see cref="EntityState"/>.IsUnchanged() method.
    /// </remarks>
    public static T ShouldBeUnchanged<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {

      return actual.AssertOnEntityStateTest("Unchanged",
        () => actual.EntityAspect.EntityState.IsUnchanged(), true,
        message, parameters);
    }

    //----------------

    /// <summary>
    /// Asserts that this entity is Added. 
    /// </summary>
    public static T ShouldBeAdded<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {

      return actual.AssertOnEntityStateTest("Added",
        () => actual.EntityAspect.EntityState.IsAdded(), true,
        message, parameters);
    }

    /// <summary>
    /// Asserts that this entity is not Added. 
    /// </summary>
    public static T ShouldNotBeAdded<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {

        return actual.AssertOnEntityStateTest("Added",
          () => actual.EntityAspect.EntityState.IsAdded(), false,
          message, parameters);
    }

    //----------------

    /// <summary>
    /// Asserts that this entity is Deleted. 
    /// </summary>
    public static T ShouldBeDeleted<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {

      return actual.AssertOnEntityStateTest("Deleted",
        () => actual.EntityAspect.EntityState.IsDeleted(), true,
        message, parameters);
    }

    /// <summary>
    /// Asserts that this entity is not Deleted. 
    /// </summary>
    public static T ShouldNotBeDeleted<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {

      return actual.AssertOnEntityStateTest("Deleted",
        () => actual.EntityAspect.EntityState.IsDeleted(), false,
        message, parameters);
    }

    //----------------

    /// <summary>
    /// Asserts that this entity is Detached. 
    /// </summary>
    public static T ShouldBeDetached<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {

      return actual.AssertOnEntityStateTest("Detached",
        () => actual.EntityAspect.EntityState.IsDetached(), true,
        message, parameters);
    }

    /// <summary>
    /// Asserts that this entity is not Detached. 
    /// </summary>
    public static T ShouldNotBeDetached<T>(
      this T actual, string message = null, params object[] parameters)
      where T : IEntity {

      return actual.AssertOnEntityStateTest("Detached",
        () => actual.EntityAspect.EntityState.IsDetached(), false,
        message, parameters);
    }
  }
}