using System;
using System.Reflection;
using IdeaBlade.EntityModel;

namespace IdeaBlade.TestFramework {

  /// <summary>
  /// Provides access to the SilverlightTest Enqueue methods on the current thread.
  /// </summary>
  public static class WaitFor {

    /// <summary>
    /// Add a Callback method into the test task queue. 
    /// Similar to the PumpMessages(...) call, with the difference being 
    /// that there is no longer a single requirement: 
    /// you can enqueue several callback methods and other test tasks, 
    /// all of which will execute before the test completes and/or the engine continues.
    /// </summary>
    /// <param name="testCallback">
    /// Void-returning delegate, anonymous delegates work fine too.
    /// </param>
    public static void Callback(Action testCallback) {
      DevForceTest.TestClassBase.EnqueueCallback(testCallback);
    }


    public static void Delay(double milliseconds) {
      DevForceTest.TestClassBase.EnqueueDelay(milliseconds);
    }
    /// <summary>
    /// Delay a minimum amount of time before continuing. Similar to a sleep call that is not a blocking call.
    /// </summary>
    /// <param name="delay">
    /// The minimum time span to wait before continuing
    /// </param>
    public static void Delay(TimeSpan delay) {
      DevForceTest.TestClassBase.EnqueueDelay(delay);
    }
    // Based on Justin Angel: http://blogs.silverlight.net/blogs/justinangel/archive/2009/02/25/silverlight-unit-testing-rhinomocks-unity-and-resharper.aspx
    public static T Event<T>(this T objectToWaitForItsEvent, string eventName) {

      EventInfo eventInfo = objectToWaitForItsEvent.GetType().GetEvent(eventName);

      bool eventRaised = false;

      if (typeof(System.Windows.RoutedEventHandler).IsAssignableFrom(eventInfo.EventHandlerType))
        eventInfo.AddEventHandler(objectToWaitForItsEvent, (System.Windows.RoutedEventHandler) delegate { eventRaised = true; });

      else if (typeof(EventHandler).IsAssignableFrom(eventInfo.EventHandlerType))
        eventInfo.AddEventHandler(objectToWaitForItsEvent, (EventHandler) delegate { eventRaised = true; });

      DevForceTest.TestClassBase.EnqueueConditional(() => eventRaised);

      return objectToWaitForItsEvent;
    }

    /// <summary>
    /// Wait until the DevForce <see cref="BaseOperation.IsCompleted"/> flag is true.
    /// Uses the <see cref="Conditional"/> method.
    /// </summary>
    /// <param name="op">Operation object returned from a DevForce async method.</param>
    public static void IsCompleted(BaseOperation op) {
      DevForceTest.TestClassBase.EnqueueConditional(() => op.IsCompleted || op.HasError || op.Cancelled);
    }
    /// <summary>
    /// Instructs the test task queue to wait until the conditional call returns True 
    /// to continue executing other test tasks and/or ending the test method.
    /// </summary>
    /// <param name="conditional">
    /// Conditional method or delegate. Test will halt until this condition returns True.
    /// </param>
    public static void Conditional(Func<bool> conditional) {
      DevForceTest.TestClassBase.EnqueueConditional(conditional);
    }
    /// <summary>
    /// Enqueue a test task which calls the <see cref="TestCompleteRightNow"/> method.
    /// </summary>
    public static void TestComplete() {
      DevForceTest.TestClassBase.EnqueueTestComplete();
    }
    /// <summary>
    /// Signal that a test is complete when using Async testing. 
    /// WARNING: If you use your own methods for completing, 
    /// such as an HtmlTimer or other threading method, 
    /// it is possible that this call will occur *AFTER* the test has timed out 
    /// when using Timeouts. 
    /// As such, be very careful as you could complete the call to the *next* test.
    /// </summary>
    public static void TestCompleteRightNow() {
      DevForceTest.TestClassBase.TestComplete();
    } 
  }
}