﻿using System;
using System.Collections.Generic;
// ReSharper disable RedundantUsingDirective
using System.ComponentModel; // keep for SL
// ReSharper restore RedundantUsingDirective
using System.Linq;
using IdeaBlade.EntityModel;
using Microsoft.Silverlight.Testing;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SimpleNorthwindModel;

// ReSharper disable InconsistentNaming
// ReSharper disable RedundantDefaultFieldInitializer
// ReSharper disable RedundantArgumentName

namespace WorkingWithEntities.EntityCacheTests
{
    [TestClass]
    public class EntityCacheTests
    {
        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
        }

        [TestMethod]
        [Tag("EntityCache")]
        public void then_no_EntityGroups_at_start()
        {
           Assert.AreEqual(0, manager.GetEntityGroups().Count()); 
        }

        [TestMethod]
        [Tag("EntityCache")]
        public void then_asking_for_EntitytGroup_creates_it()
        {
            Assert.AreEqual(0, manager.GetEntityGroups().Count(),
                "should have no grps at start"); 

            // Merely asking for the Customer grp
            var custGrp = manager.GetEntityGroup<Customer>();

            Assert.AreEqual(1, manager.GetEntityGroups().Count(),
                "should have a group after asking for it");
 
            Assert.IsNotNull(custGrp, "the grp should be non-null");
            Assert.AreEqual(0, custGrp.Count(), "the grp should be empty");
        }

        [TestMethod]
        [Tag("EntityCache")]
        public void then_have_a_Customer_EntityGroup_after_adding_Customer()
        {
            manager.AddEntity(new Customer());

            var grps = manager.GetEntityGroups();
            Assert.AreEqual(1, grps.Count(),
                "should have exactly one group");

            var custGrp = manager.GetEntityGroups().FirstOrDefault();

            Assert.IsNotNull(custGrp, "should have a grp after adding customer");
            Assert.AreSame(typeof(Customer), custGrp.EntityType, 
                "should have a Customer grp after adding customer");
            Assert.AreEqual(1, custGrp.Count(), "should have one Customer in grp");
        }


        [TestMethod]
        [Tag("EntityCache")]
        public void then_manager_has_no_EntityGroups_after_clearing()
        {
            manager.AddEntity(new Customer()); // adds Customer EntityGroup
            manager.Clear(); // clears the cache

            // Cache forgets the groups it used to have after clearing
            Assert.AreEqual(0, manager.GetEntityGroups().Count(),
                 "should have 1 grp after clear");
        }

        [TestMethod]
        [Tag("EntityCache")]
        public void then_EntityGroup_has_deleted_customer()
        {
            var cust = new Customer();
            manager.AttachEntity(cust);

            cust.EntityAspect.Delete();
            var custGrp = manager.GetEntityGroup<Customer>();

            Assert.IsTrue(custGrp.Entities.Contains(cust), "cust grp should have deleted customer");
        }

        [TestMethod]
        [Tag("EntityCache")]
        public void then_EntityGroup_CurrentEntities_lacks_deleted_customer()
        {
            var cust = new Customer();
            manager.AttachEntity(cust);

            cust.EntityAspect.Delete();
            var custGrp = manager.GetEntityGroup<Customer>();

            Assert.IsFalse(custGrp.CurrentEntities.Contains(cust), 
                "CurrentEntities should not have deleted customer");
        }

        [TestMethod]
        [Tag("EntityCache")]
        public void then_can_listen_to_EntityGroup_changes()
        {
            // Listen for changes to cache
            var cacheChanges = new List<EntityAction>();
            int cacheChangeCount = 0;
            manager.EntityChanged += (s, e) => cacheChanges.Add(e.Action);

            // Listen for changes to customers in cache
            var custChanges = new List<EntityAction>();
            int custChangeCount = 0;
            var custGrp = manager.GetEntityGroup<Customer>();
            custGrp.EntityChanged += (s, e) => custChanges.Add(e.Action);

            // Customer changes
            var cust = new Customer();
            manager.AddEntity(cust);           cacheChangeCount++; custChangeCount++;
            cust.CompanyName = "Acme";         cacheChangeCount++; custChangeCount++;
            cust.CompanyName = "Beta";         cacheChangeCount++; custChangeCount++;
            cust.EntityAspect.AcceptChanges(); cacheChangeCount++; custChangeCount++;
            cust.EntityAspect.Delete();        cacheChangeCount++; custChangeCount++;

            // Employee change
            var emp = new Employee {EmployeeID = 42};
            manager.AttachEntity(emp); cacheChangeCount++;

            Assert.AreEqual(cacheChangeCount, cacheChanges.Count(),
                 "not all cache changes were signaled");

            Assert.AreEqual(custChangeCount, custChanges.Count(), 
                "not all cust changes were signaled");

            Assert.IsTrue(cacheChangeCount > custChangeCount,
                "should have more cache changes than cust changes");
        }

        [TestMethod]
        public void can_find_a_customer_in_cache_by_key()
        {
            var cust1 = MakeUnchangedCustomer("Acme");
            var foundCust = manager.FindEntity(cust1.EntityAspect.EntityKey);
            Assert.AreSame(cust1, foundCust);
        }


        [TestMethod]
        public void can_find_a_deleted_customer_in_cache_by_key()
        {
            var cust1 = MakeUnchangedCustomer("Acme");
            cust1.EntityAspect.Delete();
            var foundCust = manager.FindEntity(
                cust1.EntityAspect.EntityKey, includeDeleted:true);
            Assert.AreSame(cust1, foundCust);
        }

        [TestMethod]
        public void can_find_a_customer_in_cache_by_query()
        {
            var cust1 = MakeUnchangedCustomer("Acme");

            var foundCust = manager
                .FindEntities(EntityState.AllButDetached)
                .OfType<Customer>()
                .Where(c => c.CompanyName == "Acme")
                .FirstOrDefault();

            Assert.AreSame(cust1, foundCust);
        }

        [TestMethod]
        public void can_transfer_entity_via_ECS()
        {
            var cust1 = MakeUnchangedCustomer("Acme");

            // Get EntityCacheState with all entities in the manager (just one)
            var ecs = manager.CacheStateManager.GetCacheState();

            // Create 2nd manager
            var manager2 = new EntityManager(shouldConnect: false);

            // Prove that it is empty
            var entities = manager2.FindEntities(EntityState.AllButDetached)
                .OfType<object>();
            Assert.AreEqual(0, entities.Count(),
                "new manager2 should be empty");

            // "Restore" with the contents of the EntityCacheState from manager
            manager2.CacheStateManager.RestoreCacheState(ecs);

            // Prove that manager2 has a customer with same ID as cust1
            var foundCust = manager2.FindEntity(cust1.EntityAspect.EntityKey);

            // But the foundCust is not the same as cust1
            // because cust1 still belongs to manager1
            Assert.AreNotSame(cust1, foundCust);

        }

        #region Test helpers

        protected Customer MakeAddedCustomer(string companyName)
        {
            var cust = new Customer
                       {
                           CompanyName = companyName,
                           Country = "Freedonia",
                       };
            manager.AddEntity(cust);
            return cust;
        }

        protected Customer MakeUnchangedCustomer(string companyName)
        {
            var cust = MakeAddedCustomer(companyName);
            cust.EntityAspect.AcceptChanges();
            return cust;
        }

        protected EntityManager manager;
        protected Customer testCustomer;

        #endregion


    }
}