﻿using System;
using System.Collections.Generic;
using IdeaBlade.EntityModel;
using System.Linq;
using Microsoft.Silverlight.Testing;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SimpleNorthwindModel;

// ReSharper disable InconsistentNaming
// ReSharper disable RedundantDefaultFieldInitializer
// ReSharper disable RedundantArgumentName

namespace WorkingWithEntities.EntityStateChangeTests
{

    [TestClass]
    public class When_toying_with_mix_of_customers : EntityStateChangeTestBase
    {
        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
            MakeTestCustomers();
        }

        [TestMethod]
        [Tag("EntityChange")]
        public void then_manager_EntityChanging_reports_changes_to_entity()
        {
            var acts = new List<EntityAction>();
            int chgCnt = 0;
            manager.EntityChanging += (s, e) => acts.Add(e.Action);
            var cust = manager.GetQuery<Customer>().First(); chgCnt++;
            cust.CompanyName = "Sally"; chgCnt++;
            cust.CompanyName = "Robin"; chgCnt++; // second mod is counted as well
            cust.EntityAspect.RejectChanges(); chgCnt++;
            manager.RemoveEntity(cust); chgCnt++;
            manager.AttachEntity(cust); chgCnt++;
            cust.EntityAspect.Delete(); chgCnt++;
            cust.EntityAspect.AcceptChanges(); chgCnt++;

            Assert.AreEqual(chgCnt,acts.Count, "changes detected should equal # of changes");
        }

        [TestMethod]
        [Tag("EntityState")]
        public void testCustomers_are_in_expected_states()
        {
            AssertIsAdded(addedCustomer);
            AssertIsUnchanged(unchangedCustomer);
            AssertIsModified(modifiedCustomer);
            AssertIsDeleted(deletedCustomer);
        }

        [TestMethod]
        [Tag("EntityState")]
        public void change_to_unchanged_entity_isModified()
        {
            unchangedCustomer.CompanyName = "abracadabra";
            AssertIsModified(unchangedCustomer);
        }

        [TestMethod]
        [Tag("EntityState")]
        public void then_delete_of_added_entity_isDetached()
        {
            addedCustomer.EntityAspect.Delete();
            AssertIsDetached(addedCustomer);
        }

        [TestMethod]
        [Tag("EntityState")]
        public void undo_of_added_entity_isDetached()
        {
            addedCustomer.EntityAspect.RejectChanges();
            AssertIsDetached(addedCustomer);
        }

        [TestMethod]
        [Tag("EntityState")]
        public void isChanged_of_detached_unchanged_entity_is_true()
        {
            // IsChanged and HasChanges() are defined in terms of EntityState
            // An entity IsChanged if it's EntityState is other than Unchanged.
            unchangedCustomer.EntityAspect.RemoveFromManager();
            Assert.IsTrue(addedCustomer.EntityAspect.IsChanged);
        }

        [TestMethod]
        [Tag("EntityState")]
        public void delete_of_existing_entity_isDeleted()
        {
            unchangedCustomer.EntityAspect.Delete();
            AssertIsDeleted(unchangedCustomer, "unchanged should be deleted");

            modifiedCustomer.EntityAspect.Delete();
            AssertIsDeleted(modifiedCustomer, "modified should be deleted");
        }

        [TestMethod]
        [Tag("EntityState")]
        public void undo_of_modified_entity_isUnchanged()
        {
            modifiedCustomer.EntityAspect.RejectChanges();
            AssertIsUnchanged(modifiedCustomer);
        }

    }

    [TestClass]
    public class When_setting_navigation_properties : EntityStateChangeTestBase
    {
        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
            unchangedCustomer = MakeUnchangedCustomer("Test Company");           
            newOrder = new Order();
        }

        private Order newOrder;

        [TestMethod]
        [Tag("EntityState")]
        public void assigning_ref_nav_adds_assignee_to_manager()
        {
            AssertIsDetached(newOrder, "newOrder should be detached at start.");

            newOrder.Customer = unchangedCustomer;

            AssertIsAdded(newOrder, "newOrder should be added after nav assignment.");

            Assert.AreSame(
                newOrder.EntityAspect.EntityManager,
                unchangedCustomer.EntityAspect.EntityManager,
                "EntityManagers should be the same");

            AssertIsUnchanged(unchangedCustomer,
            "adding to unchangedCustomer's orders shouldn't modify the customer itself");
        }

        [TestMethod]
        [Tag("EntityState")]
        public void assigning_ref_nav_id_wont_add_assignee_to_manager()
        {
            // Lesson: if you DO NOT want the new entity to be added
            // (e.g., you are making a test or design entity) 
            // then DO NOT set one of its reference navigation properties.
            // Set the corresponding FK id property instead.

            newOrder.CustomerID = unchangedCustomer.CustomerID;

            AssertIsDetached(newOrder, "newOrder should be detached after FK assignment.");

            manager.AttachEntity(newOrder);

            AssertIsUnchanged(newOrder, "newOrder should be unchanged after attach.");

            var orders = unchangedCustomer.Orders;
            Assert.IsTrue(orders.Contains(newOrder),
                          "newOrder should belong to unchangedCustomer.");
        }

        [TestMethod]
        [Tag("EntityState")]
        public void assigning_ref_nav_of_detached_cust_wont_add_order_to_manager()
        {
            var newCust = new Customer { CustomerID = Guid.NewGuid() };

            newOrder.Customer = newCust; // NEITHER has an EntityManager!

            AssertIsDetached(newCust, "newCust should be detached.");
            AssertIsDetached(newOrder, "newOrder should be detached.");

            manager.AttachEntity(newCust);  // This works because GUID Id
            manager.AttachEntity(newOrder); // Danger: Id always 0 because auto-inc Id !!

            AssertIsUnchanged(newCust, "newCust should be unchanged after attach.");
            AssertIsUnchanged(newOrder, "newOrder should be unchanged after attach.");

            var orders = newCust.Orders;
            Assert.IsTrue(orders.Contains(newOrder),
                          "newOrder should belong to unchangedCustomer.");
        }

        [TestMethod]
        [Tag("EntityState")]
        public void adding_detached_entity_to_existing_entity_collection_adds_to_manager()
        {
  
            unchangedCustomer.Orders.Add(newOrder);

            AssertIsAdded(newOrder, "newOrder should be added after nav assignment.");

            Assert.AreSame(
                newOrder.EntityAspect.EntityManager,
                unchangedCustomer.EntityAspect.EntityManager,
                "EntityManagers should be the same");

            AssertIsUnchanged(unchangedCustomer,
            "adding to unchangedCustomer's orders shouldn't modify the customer itself");
        }
    }

    [TestClass]
    public class When_using_EntityState_Set_methods : EntityStateChangeTestBase
    {
        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
            MakeTestCustomers();
        }

        [TestMethod]
        [Tag("EntityState")]
        public void unchanged_becomes_Added_when_call_SetAdded()
        {
            unchangedCustomer.EntityAspect.SetAdded();
            AssertIsAdded(unchangedCustomer);
        }

        [TestMethod]
        [Tag("EntityState")]
        public void unchanged_becomes_Modified_when_call_SetModified()
        {
            unchangedCustomer.EntityAspect.SetModified();
            AssertIsModified(unchangedCustomer);
        }

        [TestMethod]
        [Tag("EntityState")]
        public void can_chain_SetAdded_SetModified_Undo()
        {
            unchangedCustomer.EntityAspect.SetAdded();
            AssertIsAdded(unchangedCustomer, " after SetAdded");
            unchangedCustomer.EntityAspect.SetModified();
            AssertIsModified(unchangedCustomer, " after SetModified");
            unchangedCustomer.EntityAspect.RejectChanges();
            AssertIsUnchanged(unchangedCustomer, " after Undo");
        }
    }

    [TestClass]
    public class When_looking_in_entity_cache : EntityStateChangeTestBase
    {
        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
            MakeTestCustomers();
        }

        [TestMethod]
        [Tag("EntityState")]
        public void can_find_all_test_customers()
        {
            var entities = manager
                .FindEntities(EntityState.AllButDetached)
                .OfType<Entity>();
            
            Assert.IsTrue(entities.Contains(addedCustomer), "missing added");
            Assert.IsTrue(entities.Contains(unchangedCustomer), "missing unchanged");
            Assert.IsTrue(entities.Contains(modifiedCustomer), "missing modified");
            Assert.IsTrue(entities.Contains(deletedCustomer), "missing deleted");
            Assert.AreEqual(4, entities.Count(), "Should have exactly 4 in cache");
        }


        [TestMethod]
        [Tag("EntityState")]
        public void can_find_all_changed_test_customers()
        {
            var entities = manager
                .FindEntities(EntityState.AnyAddedModifiedOrDeleted)
                .OfType<Entity>();

            Assert.IsFalse(entities.Contains(unchangedCustomer), "found unchanged");

            Assert.IsTrue(entities.Contains(addedCustomer), "missing added");
            Assert.IsTrue(entities.Contains(modifiedCustomer), "missing modified");
            Assert.IsTrue(entities.Contains(deletedCustomer), "missing deleted");
            Assert.AreEqual(3, entities.Count(), "Should have exactly 3 changed in cache");
        }

        [TestMethod]
        [Tag("EntityState")]
        public void dont_find_detached_in_cache()
        {
            unchangedCustomer.EntityAspect.RemoveFromManager();

            var entities = manager
                .FindEntities(EntityState.AnyAddedModifiedOrDeleted)
                .OfType<Entity>();

            Assert.IsFalse(entities.Contains(unchangedCustomer), "found unchanged");
        } 
    }

    [TestClass]
    public class When_combining_EntityState_enums : EntityStateChangeTestBase
    {
        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
            MakeTestCustomers();
        }

        [TestMethod]
        [Tag("EntityState")]
        public void can_combine_EntityState_enums_to_check_entity_state()
        {
            var currentState = unchangedCustomer.EntityAspect.EntityState;
            if ((currentState & (EntityState.Added | EntityState.Modified)) > 0)
            {
                Assert.Fail("Shouldn't be added or modified");
            }
        }

        [TestMethod]
        [Tag("EntityState")]
        public void can_use_EntityState_extensions_to_check_entity_state()
        {
            var currentState = unchangedCustomer.EntityAspect.EntityState;
            if (currentState.IsAddedOrModified())
            {
                Assert.Fail("Shouldn't be added or modified");
            }
        }

        [TestMethod]
        [Tag("EntityState")]
        public void can_combine_EntityState_enums_to_find_changed_entities_in_Cache()
        {
            const EntityState desiredStates = EntityState.Added | 
                                              EntityState.Modified | 
                                              EntityState.Deleted;
            var changedEntities = manager.FindEntities<Customer>(desiredStates);
            Assert.AreEqual(3, changedEntities.Count());
        }

        [TestMethod]
        [Tag("EntityState")]
        public void can_use_predefined_combo_enum_to_find_changed_entities_in_Cache()
        {
            var changedEntities = 
                manager.FindEntities<Customer>(EntityState.AnyAddedModifiedOrDeleted);
            Assert.AreEqual(3, changedEntities.Count());
        }
    }

    public abstract class EntityStateChangeTestBase {

        protected EntityManager manager;

        #region EntityStateAsserts

        protected void AssertIsAdded(IEntity entity, string message = null)
        {
            Assert.AreEqual(EntityState.Added, entity.EntityAspect.EntityState, message);
        }
        protected void AssertIsDeleted(IEntity entity, string message = null)
        {
            Assert.AreEqual(EntityState.Deleted, entity.EntityAspect.EntityState, message);
        }
        protected void AssertIsDetached(IEntity entity, string message=null)
        {
            Assert.AreEqual(EntityState.Detached, entity.EntityAspect.EntityState, message);
        }
        protected void AssertIsModified(IEntity entity, string message = null)
        {
            Assert.AreEqual(EntityState.Modified, entity.EntityAspect.EntityState, message);
        }
        protected void AssertIsUnchanged(IEntity entity, string message = null)
        {
            Assert.AreEqual(EntityState.Unchanged, entity.EntityAspect.EntityState, message);
        }

        #endregion

        #region MakeTestCustomers

        protected void MakeTestCustomers()
        {
            addedCustomer = MakeAddedCustomer("Added Company");
            unchangedCustomer = MakeUnchangedCustomer("Unchanged Company");
            modifiedCustomer = MakeModifiedCustomer("Modified Company");
            deletedCustomer = MakeDeletedCustomer("Deleted Company");
        }

        protected Customer addedCustomer;
        protected Customer unchangedCustomer;
        protected Customer modifiedCustomer;
        protected Customer deletedCustomer;

        protected Customer MakeAddedCustomer(string companyName)
        {
            var cust = new Customer
                       {
                           CompanyName = companyName,
                           Country = "Freedonia",
                       };
            manager.AddEntity(cust);
            return cust;
        }

        protected Customer MakeUnchangedCustomer(string companyName)
        {
            var cust = MakeAddedCustomer(companyName);
            cust.EntityAspect.AcceptChanges();
            return cust;
        }

        protected Customer MakeModifiedCustomer(string companyName)
        {
            var cust = MakeUnchangedCustomer(companyName);
            cust.EntityAspect.SetModified();
            return cust;
        }

        protected Customer MakeDeletedCustomer(string companyName)
        {
            var cust = MakeUnchangedCustomer(companyName);
            cust.EntityAspect.Delete();
            return cust;
        }

        #endregion

    }
}