﻿using System;
using System.ComponentModel; // keep for SL
using IdeaBlade.EntityModel;
using Microsoft.Silverlight.Testing; 
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SimpleNorthwindModel;

// ReSharper disable InconsistentNaming
// ReSharper disable RedundantDefaultFieldInitializer
// ReSharper disable RedundantArgumentName

namespace WorkingWithEntities.EntityVersionTests
{

    [TestClass]
    public class When_start_with_unchanged_test_customer : EntityVersionTests
    {
        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
            testCustomer = new Customer
            {
                CustomerID = Guid.NewGuid(),
                CompanyName = testCompanyName,
            };
            AttachTestCustomerAsUnchanged();
        }

        [TestMethod]
        [Tag("Version")]
        public void then_testCustomer_isUnchanged_at_start()
        {
            Assert.IsTrue(TestCustomerEntityState.IsUnchanged());
        }

        [TestMethod]
        [Tag("Version")]
        public void then_orig_equals_cur_at_start()
        {
            Assert.AreEqual(testCompanyName, GetOriginalName(),
                           "testCompanyName should match orig");
        }

        [TestMethod]
        [Tag("Version")]
        public void can_access_property_if_detached()
        {
            testCustomer.EntityAspect.RemoveFromManager();
            Assert.IsTrue(TestCustomerEntityState.IsDetached());
            Assert.AreEqual(testCompanyName, testCustomer.CompanyName);
        }

        [TestMethod]
        [Tag("Version")]
        public void then_orig_is_defined_after_detach()
        {
            testCustomer.EntityAspect.RemoveFromManager();
            Assert.IsTrue(
                TestCustomerEntityState.IsDetached(),
                "Should be detached");

            Assert.AreEqual(testCompanyName, GetOriginalName(),
                           "testCompanyName should match orig");
        }

        [TestMethod]
        [Tag("Version")]
        public void then_cur_differs_from_orig_if_modified()
        {
            const string newName = "New Name";
            testCustomer.CompanyName = newName;
            Assert.IsTrue(
                TestCustomerEntityState.IsModified(),
                "Should be modified");

            Assert.AreEqual(GetName(), GetCurrentName(),
                           "property should match cur");

            Assert.AreEqual(newName, GetCurrentName(),
                           "newName should match cur");

            Assert.AreEqual(testCompanyName, GetOriginalName(),
                           "testCompanyName should match orig");
        }

        [TestMethod]
        [Tag("Version")]
        public void then_cur_equals_orig_after_undo()
        {
            var origName = testCustomer.CompanyName;
            const string newName = "New Name";

            testCustomer.CompanyName = newName;

            Assert.IsTrue(
                TestCustomerEntityState.IsModified(),
                "Should be modified after setting name");
    
            testCustomer.EntityAspect.RejectChanges();

            Assert.IsTrue(
                TestCustomerEntityState.IsUnchanged(),
                "Should be unchanged after undo");

            Assert.AreEqual(origName, GetName(),
                           "property should match orig");

            Assert.AreEqual(origName, GetCurrentName(),
                           "origName should match cur");

            Assert.AreEqual(origName, GetOriginalName(),
                           "origName should match orig");
        }

        [TestMethod]
        [Tag("Version")]
        public void then_default_equals_orig_after_delete()
        {
            var origName = testCustomer.CompanyName;
            const string newName = "New Name";

            testCustomer.CompanyName = newName;

            Assert.IsTrue(
                TestCustomerEntityState.IsModified(),
                "Should be modified after setting name");

            testCustomer.EntityAspect.Delete();

            Assert.IsTrue(
                TestCustomerEntityState.IsDeleted(),
                "Should be deleted after delete");

            // Default maps to Original Version
            Assert.AreEqual(origName, GetDefaultName(),
                           "origName should match default");

            Assert.AreEqual(origName, GetName(),
                           "origName should match property");

            // Cur version is untouched
            Assert.AreEqual(newName, GetCurrentName(),
                           "newName should match cur");

            Assert.AreEqual(origName, GetOriginalName(),
                           "origName should match orig");
        }

        [TestMethod]
        [Tag("Version")]
        public void then_put_it_through_the_wringer()
        {
            // testCustomer is unchanged as we begin

            var origName = testCustomer.CompanyName;
            const string newName = "New Name";
            const string proposedName = "Proposed Name";

            // Change the current value.
            testCustomer.CompanyName = newName;

            Assert.IsTrue(
                TestCustomerEntityState.IsModified(),
                "Should be modified");

            Assert.AreEqual(newName, GetName(),
                           "newName should match the property getter value before edit");

            Assert.AreEqual(newName, GetDefaultName(),
                           "newName should match the default value before edit");

            // before edit session, proposed is current
            Assert.AreEqual(newName, GetProposedName(),
                           "newName should match the proposed value before edit");

            Assert.AreEqual(newName, GetCurrentName(),
                           "newName should match the current value before edit");

            Assert.AreEqual(origName, GetOriginalName(),
                           "origName should match the original value before edit");

            // Open IEditableObject editing session
            var ieo = testCustomer as IEditableObject;
            ieo.BeginEdit();

            // set tentative value within editing session
            testCustomer.CompanyName = proposedName;

            Assert.AreEqual(proposedName, GetName(),
                           "proposedName should match the property getter value during edit");

            Assert.AreEqual(proposedName, GetDefaultName(),
                           "proposedName should match the default value during edit");

            Assert.AreEqual(proposedName, GetProposedName(),
                           "proposedName should match the proposed value during edit");

            Assert.AreNotEqual(proposedName, GetCurrentName(),
                           "proposedName should not match the current value during edit");

            Assert.AreEqual(origName, GetOriginalName(),
                           "origName should match the original value during edit");

            // Cancel IEditableObject editing session
            ieo.CancelEdit();

            // After cancel, current should revert

            Assert.AreEqual(newName, GetName(),
                           "newName should match the property getter value after cancel");

            Assert.AreEqual(newName, GetDefaultName(),
                           "newName should match the default value after cancel");

            // Notice that proposed reverted too
            Assert.AreEqual(newName, GetProposedName(),
                           "newName should match the proposed value after cancel");

            Assert.AreEqual(newName, GetCurrentName(),
                           "newName should match the current value after cancel");

            Assert.AreEqual(origName, GetOriginalName(),
                           "origName should match the original value after cancel");

            // Now detach 
            testCustomer.EntityAspect.RemoveFromManager();

            Assert.AreEqual(newName, GetName(),
                           "newName should match the property getter value of detached");

            Assert.AreEqual(newName, GetDefaultName(),
                           "newName should match the default value of detached");

            Assert.AreEqual(newName, GetProposedName(),
                           "newName should match the proposed value of detached");

            Assert.AreEqual(newName, GetCurrentName(),
                           "newName should match the current value of detached");

            Assert.AreEqual(origName, GetOriginalName(),
                           "origName should match the original value of detached");

            // Now re-attach AS MODIFIED
            manager.AttachEntity(testCustomer, EntityState.Modified);
            Assert.IsTrue(
                TestCustomerEntityState.IsModified(),
                "Should be modified after re-attachment as modified");

            Assert.AreEqual(newName, GetName(),
                           "newName should match the property getter value of re-attached");

            Assert.AreEqual(newName, GetDefaultName(),
                           "newName should match the default value of re-attached");

            Assert.AreEqual(newName, GetProposedName(),
                           "newName should match the proposed value of re-attached");

            Assert.AreEqual(newName, GetCurrentName(),
                           "newName should match the current value of re-attached");

            // N.B. The original is restored when re-attached
            Assert.AreEqual(origName, GetOriginalName(),
                           "origName should match the original value of re-attached");
        }

    }

    [TestClass]
    public class When_start_with_new_test_customer : EntityVersionTests
    {
        // Added entities don't have original
        // If you ask for the original of an added entity
        // you get the "default"

        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
            testCustomer = new Customer
            {
                CompanyName = testCompanyName,
                Country = "Freedonia",
            };
            manager.AddEntity(testCustomer);
        }

        [TestMethod]
        [Tag("Version")]
        public void then_testCustomer_isAdded_at_start()
        {
            Assert.IsTrue(TestCustomerEntityState.IsAdded());
        }

        [TestMethod]
        [Tag("Version")]
        public void then_default_equals_cur_after_delete()
        {
            var initialName = GetName();

            Assert.IsTrue(
                TestCustomerEntityState.IsAdded(),
                "Should be added at start");

            testCustomer.EntityAspect.Delete();

            // New entities are detached after they are deleted
            // No point in saving them as there should be no
            // corresponding database object to delete.
            Assert.IsTrue(
                TestCustomerEntityState.IsDetached(),
                "Should be detached after delete");

            // Default maps to the pre-delete current value
            Assert.AreEqual(initialName, GetDefaultName(),
                           "initialName should match default");

            Assert.AreEqual(initialName, GetName(),
                           "initialName should match property");

            Assert.AreEqual(initialName, GetCurrentName(),
                           "initialName should match cur");

        }

    }

    #region EntityVersionTests base class

    public abstract class EntityVersionTests
    {

        protected Customer testCustomer;
        protected const string testCompanyName = "Test Company";
        protected EntityManager manager;

        protected void AttachTestCustomerAsUnchanged()
        {
            manager.AttachEntity(testCustomer);
            testCustomer.EntityAspect.SetModified();
            testCustomer.EntityAspect.AcceptChanges();
        }

        protected EntityState TestCustomerEntityState
        {
            get { return testCustomer.EntityAspect.EntityState; }
        }

        #region GetName in all versions

        protected string GetName()
        {
            return testCustomer.CompanyName;
        }

        protected string GetCurrentName()
        {
            return (string) testCustomer.EntityAspect
                                .GetValue("CompanyName", EntityVersion.Current);
        }

        protected string GetOriginalName()
        {
            return (string) testCustomer.EntityAspect
                                .GetValue("CompanyName", EntityVersion.Original);
        }

        protected string GetProposedName()
        {
            return (string) testCustomer.EntityAspect
                                .GetValue("CompanyName", EntityVersion.Proposed);
        }

        protected string GetDefaultName()
        {
            return (string) testCustomer.EntityAspect
                                .GetValue("CompanyName", EntityVersion.Default);
        }

        #endregion      
    }

    #endregion

}