Imports System.Collections

Namespace IdeaBlade.EntityModel

  #Region "Non-generic EntityQueryOperation Extensions"

  Public Module EntityQueryOperationExtensions

	Public OperationWrapperFactory As Func(Of EntityQueryOperation, IEntityQueryOperationWrapper) = Function(op) New DefaultEntityQueryOperationWrapper(op)

	Private Function MakeWrapper(ByVal op As EntityQueryOperation) As IEntityQueryOperationWrapper
	  Return OperationWrapperFactory(op)
	End Function

	<System.Runtime.CompilerServices.Extension> _
	Public Function OnCancelled(ByVal op As EntityQueryOperation, ByVal action As Action) As IEntityQueryOperationWrapper
	  Return MakeWrapper(op).OnCancelled(action)
	End Function

	''' <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
	<System.Runtime.CompilerServices.Extension> _
	Public Sub OnCompleted(ByVal op As EntityQueryOperation, ByVal action As Action(Of EntityQueriedEventArgs))
	  MakeWrapper(op).OnCompleted(action)
	End Sub

	<System.Runtime.CompilerServices.Extension> _
	Public Function OnError(ByVal op As EntityQueryOperation, ByVal action As Action(Of Exception)) As IEntityQueryOperationWrapper
	  Return MakeWrapper(op).OnError(action)
	End Function

	<System.Runtime.CompilerServices.Extension> _
	Public Function OnError(ByVal op As EntityQueryOperation, Optional ByVal failureMessageTemplate As String = Nothing) As IEntityQueryOperationWrapper
	  Return MakeWrapper(op).OnError(failureMessageTemplate)
	End Function

	<System.Runtime.CompilerServices.Extension> _
	Public Function Into(ByVal op As EntityQueryOperation, ByVal targetCollection As IList) As IEntityQueryOperationWrapper
	  Return MakeWrapper(op).Into(targetCollection)
	End Function


	<System.Runtime.CompilerServices.Extension> _
	Public Function OnSuccess(ByVal op As EntityQueryOperation, ByVal action As Action(Of IEnumerable)) As IEntityQueryOperationWrapper
	  Return MakeWrapper(op).OnSuccess(action)
	End Function
  End Module

  Public Interface IEntityQueryOperationWrapper

	Function Into(ByVal targetCollection As IList) As IEntityQueryOperationWrapper
	Function OnSuccess(ByVal action As Action(Of IEnumerable)) As IEntityQueryOperationWrapper
	Function OnCancelled(ByVal action As Action) As IEntityQueryOperationWrapper
	''' <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
	Sub OnCompleted(ByVal action As Action(Of EntityQueriedEventArgs))
	Function OnError(ByVal action As Action(Of Exception)) As IEntityQueryOperationWrapper
	Function OnError(Optional ByVal failureMessageTemplate As String = Nothing) As IEntityQueryOperationWrapper

  End Interface

  Public Class DefaultEntityQueryOperationWrapper
	  Implements IEntityQueryOperationWrapper

	Public Sub New(ByVal op As EntityQueryOperation)
	  Operation = op
	  OnCompletedHandler = New OperationOnCompletedHandler(Of EntityQueryOperation, EntityQueriedEventArgs)(op)
	  AddHandler op.Completed, AddressOf OnCompletedHandler.CompletedHandler
	End Sub

	Private privateOperation As EntityQueryOperation
	Public Property Operation() As EntityQueryOperation
		Get
			Return privateOperation
		End Get
		Private Set(ByVal value As EntityQueryOperation)
			privateOperation = value
		End Set
	End Property

	Protected Property OnCompletedHandler() As OperationOnCompletedHandler(Of EntityQueryOperation, EntityQueriedEventArgs)

	Public Function OnCancelled(ByVal action As Action) As IEntityQueryOperationWrapper Implements IEntityQueryOperationWrapper.OnCancelled
	  OnCompletedHandler.OnCancelled(action)
	  Return Me
	End Function

	Public Sub OnCompleted(ByVal action As Action(Of EntityQueriedEventArgs)) Implements IEntityQueryOperationWrapper.OnCompleted
	  OnCompletedHandler.OnCompleted(action)
	End Sub

	Public Function OnError(ByVal action As Action(Of Exception)) As IEntityQueryOperationWrapper Implements IEntityQueryOperationWrapper.OnError
	  OnCompletedHandler.OnError(action)
	  Return Me
	End Function

	Public Function OnError(Optional ByVal failureMessageTemplate As String = Nothing) As IEntityQueryOperationWrapper Implements IEntityQueryOperationWrapper.OnError
	  OnCompletedHandler.OnError(MakeFailureMessageTemplate(failureMessageTemplate))
	  Return Me
	End Function

	Private Function MakeFailureMessageTemplate(ByVal failureMessageTemplate As String) As String
	  If String.IsNullOrWhiteSpace(failureMessageTemplate) Then
		failureMessageTemplate = "EntityQuery failed with exception ""{0}""."
	  End If
	  Return failureMessageTemplate
	End Function

	Public Function Into(ByVal targetCollection As IList) As IEntityQueryOperationWrapper Implements IEntityQueryOperationWrapper.Into
	  OnCompletedHandler.OnSuccess(Sub(args)
		  Dim results = args.Results
		  For Each item In results
			  targetCollection.Add(item)
		  Next item
	  End Sub)
	  Return Me
	End Function

	Public Function OnSuccess(ByVal action As Action(Of IEnumerable)) As IEntityQueryOperationWrapper Implements IEntityQueryOperationWrapper.OnSuccess
      OnCompletedHandler.OnSuccess(Sub(args) action(args.Results))
	  Return Me
	End Function

  End Class

  #End Region

  #Region "EntityQueryOperation<T> Extensions"

  Public Module EntityQueryOperationOfTExtensions

	' ToDo: Figure out how to substitute a custom MakeWrapper implementation.
	' Hurdle is can't creat extension method in a generic static class

	Private Function MakeWrapper(Of T)(ByVal op As EntityQueryOperation(Of T)) As IEntityQueryOperationWrapper(Of T)
	  Return New DefaultEntityQueryOperationWrapper(Of T)(op)
	End Function

	<System.Runtime.CompilerServices.Extension> _
	Public Function OnCancelled(Of T)(ByVal op As EntityQueryOperation(Of T), ByVal action As Action) As IEntityQueryOperationWrapper(Of T)
	  Return MakeWrapper(op).OnCancelled(action)
	End Function

	''' <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
	<System.Runtime.CompilerServices.Extension> _
	Public Sub OnCompleted(Of T)(ByVal op As EntityQueryOperation(Of T), ByVal action As Action(Of EntityQueriedEventArgs(Of T)))
	  MakeWrapper(op).OnCompleted(action)
	End Sub

	<System.Runtime.CompilerServices.Extension> _
	Public Function OnError(Of T)(ByVal op As EntityQueryOperation(Of T), ByVal action As Action(Of Exception)) As IEntityQueryOperationWrapper(Of T)
	  Return MakeWrapper(op).OnError(action)
	End Function

	<System.Runtime.CompilerServices.Extension> _
	Public Function OnError(Of T)(ByVal op As EntityQueryOperation(Of T), Optional ByVal failureMessageTemplate As String = Nothing) As IEntityQueryOperationWrapper(Of T)
	  Return MakeWrapper(op).OnError(failureMessageTemplate)
	End Function

	<System.Runtime.CompilerServices.Extension> _
	Public Function Into(Of T)(ByVal op As EntityQueryOperation(Of T), ByVal targetCollection As IList) As IEntityQueryOperationWrapper(Of T)
	  Return MakeWrapper(op).Into(targetCollection)
	End Function

	<System.Runtime.CompilerServices.Extension> _
	Public Function OnSuccess(Of T)(ByVal op As EntityQueryOperation(Of T), ByVal action As Action(Of IEnumerable(Of T))) As IEntityQueryOperationWrapper(Of T)
	  Return MakeWrapper(op).OnSuccess(action)
	End Function
  End Module

  Public Interface IEntityQueryOperationWrapper(Of T)

	Function Into(ByVal targetCollection As IList) As IEntityQueryOperationWrapper(Of T)
	Function OnSuccess(ByVal action As Action(Of IEnumerable(Of T))) As IEntityQueryOperationWrapper(Of T)
	Function OnCancelled(ByVal action As Action) As IEntityQueryOperationWrapper(Of T)
	''' <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
	Sub OnCompleted(ByVal action As Action(Of EntityQueriedEventArgs(Of T)))
	Function OnError(ByVal action As Action(Of Exception)) As IEntityQueryOperationWrapper(Of T)
	Function OnError(Optional ByVal failureMessageTemplate As String = Nothing) As IEntityQueryOperationWrapper(Of T)

  End Interface

  Public Class DefaultEntityQueryOperationWrapper(Of T)
	  Implements IEntityQueryOperationWrapper(Of T)

	Public Sub New(ByVal op As EntityQueryOperation(Of T))
	  Operation = op
	  OnCompletedHandler = New OperationOnCompletedHandler(Of EntityQueryOperation(Of T), EntityQueriedEventArgs(Of T))(op)
	  AddHandler op.Completed, AddressOf OnCompletedHandler.CompletedHandler
	End Sub

	Private privateOperation As EntityQueryOperation(Of T)
	Public Property Operation() As EntityQueryOperation(Of T)
		Get
			Return privateOperation
		End Get
		Private Set(ByVal value As EntityQueryOperation(Of T))
			privateOperation = value
		End Set
	End Property

	Protected Property OnCompletedHandler() As OperationOnCompletedHandler(Of EntityQueryOperation(Of T), EntityQueriedEventArgs(Of T))

	Public Function OnCancelled(ByVal action As Action) As IEntityQueryOperationWrapper(Of T) Implements IEntityQueryOperationWrapper(Of T).OnCancelled
	  OnCompletedHandler.OnCancelled(action)
	  Return Me
	End Function

	Public Sub OnCompleted(ByVal action As Action(Of EntityQueriedEventArgs(Of T))) Implements IEntityQueryOperationWrapper(Of T).OnCompleted
	  OnCompletedHandler.OnCompleted(action)
	End Sub

	Public Function OnError(ByVal action As Action(Of Exception)) As IEntityQueryOperationWrapper(Of T) Implements IEntityQueryOperationWrapper(Of T).OnError
	  OnCompletedHandler.OnError(action)
	  Return Me
	End Function

	Public Function OnError(Optional ByVal failureMessageTemplate As String = Nothing) As IEntityQueryOperationWrapper(Of T) Implements IEntityQueryOperationWrapper(Of T).OnError
	  OnCompletedHandler.OnError(MakeFailureMessageTemplate(failureMessageTemplate))
	  Return Me
	End Function

	Private Function MakeFailureMessageTemplate(ByVal failureMessageTemplate As String) As String
	  If String.IsNullOrWhiteSpace(failureMessageTemplate) Then
		Dim queryTypeName = (CType(Operation.EntityQuery, IQueryable)).ElementType.Name
		failureMessageTemplate = "EntityQuery<" & queryTypeName & "> failed with exception ""{0}""."
	  End If
	  Return failureMessageTemplate
	End Function

	Public Function Into(ByVal targetCollection As IList) As IEntityQueryOperationWrapper(Of T) Implements IEntityQueryOperationWrapper(Of T).Into
	  OnCompletedHandler.OnSuccess(Sub(args)
		  Dim results = args.Results
		  For Each item In results
			  targetCollection.Add(item)
		  Next item
	  End Sub)
	  Return Me
	End Function

	Public Function OnSuccess(ByVal action As Action(Of IEnumerable(Of T))) As IEntityQueryOperationWrapper(Of T) Implements IEntityQueryOperationWrapper(Of T).OnSuccess
      OnCompletedHandler.OnSuccess(Sub(args) action(args.Results))
	  Return Me
	End Function

  End Class

  #End Region

End Namespace