Imports System.ComponentModel
' ReSharper disable InconsistentNaming

Namespace IdeaBlade.EntityModel

  Public Class PropertyChangedCoroutineOperation
	  Implements INotifyCompleted

	''' <summary>
	''' Wait upon a <see cref="INotifyPropertyChanged.PropertyChanged"/>
	''' event for the specified property.
	''' </summary>
	''' <param name="inpc">
	''' Object that implements <see cref="INotifyPropertyChanged.PropertyChanged"/>.
	''' </param>
	''' <param name="propertyName">
	''' Name of the property to watch. Null or empty means watch for any
	''' <see cref="INotifyPropertyChanged.PropertyChanged"/> event.
	''' </param>
	''' <param name="matchEmptyPropertyName">
	''' If true, will also complete if the PropertyName in the
	''' <see cref="INotifyPropertyChanged.PropertyChanged"/> EventArgs is
	''' null or empty.
	''' </param>
	Public Sub New(ByVal inpc As INotifyPropertyChanged, Optional ByVal propertyName As String = Nothing, Optional ByVal matchEmptyPropertyName As Boolean = False)
		Me.New(inpc, Nothing, propertyName, matchEmptyPropertyName)
	End Sub

	''' <summary>
	''' Wait upon a <see cref="INotifyPropertyChanged.PropertyChanged"/>
	''' event for the specified property and for the property value to
	''' satisfy the <see cref="propertyValueMatchPredicate"/> test.
	''' </summary>
	''' <param name="inpc">
	''' Object that implements <see cref="INotifyPropertyChanged.PropertyChanged"/>.
	''' </param>
	''' <param name="propertyValueMatchPredicate">
	''' Function that reports true if the the operation should complete.
	''' Typically tests for a specific value of the watched property.
	''' Completes immediately if the function returns true from the start.
	''' </param>
	''' <param name="propertyName">
	''' Name of the property to watch. Null or empty means watch for any
	''' <see cref="INotifyPropertyChanged.PropertyChanged"/> event.
	''' </param>
	''' <param name="matchEmptyPropertyName">
	''' If true, will also complete if the PropertyName in the
	''' <see cref="INotifyPropertyChanged.PropertyChanged"/> EventArgs is
	''' null or empty.
	''' </param>
	''' <remarks>
	''' The <see cref="propertyValueMatchPredicate"/>, if non-null,
	''' is evaluated during construction and, if true, the operation completes
	''' immediately without bothering to wait on PropertyChanged.
	''' </remarks>
	Public Sub New(ByVal inpc As INotifyPropertyChanged, ByVal propertyValueMatchPredicate As Func(Of Boolean), Optional ByVal propertyName As String = Nothing, Optional ByVal matchEmptyPropertyName As Boolean = False)

	  CompletedEventArgs = New PropertyChangedCoroutineCompletedEventArgs(inpc, propertyValueMatchPredicate, propertyName, matchEmptyPropertyName)

	  AddHandler CompletedEventArgs.Completed, Sub(s, args) RaiseCompleted()
	End Sub

	Public ReadOnly Property ActualPropertyName() As String
		Get
			Return CompletedEventArgs.ActualPropertyName
		End Get
	End Property
	Public ReadOnly Property ExpectedPropertyName() As String
		Get
			Return CompletedEventArgs.ExpectedPropertyName
		End Get
	End Property
	Public ReadOnly Property IsCompleted() As Boolean
		Get
			Return CompletedEventArgs.IsCompleted
		End Get
	End Property
	Public ReadOnly Property MatchEmptyPropertyName() As Boolean
		Get
			Return CompletedEventArgs.MatchEmptyPropertyName
		End Get
	End Property

	Private ReadOnly CompletedEventArgs As PropertyChangedCoroutineCompletedEventArgs

    Public Overridable Sub WhenCompleted(ByVal completedAction As Action(Of INotifyCompletedArgs)) Implements INotifyCompleted.WhenCompleted
      If Nothing Is completedAction Then
        Return
      End If
      If IsCompleted Then
        completedAction(CompletedEventArgs)
        Return
      End If
      _notifyCompletedActions = CType(System.Delegate.Combine(_notifyCompletedActions, completedAction), Action(Of INotifyCompletedArgs))
    End Sub

	Protected Sub RaiseCompleted()
	  CallCompletedActions()
      If Nothing Is CompletedEventArgs Then
        Return
      End If
      RaiseEvent Completed(Me, CompletedEventArgs)
	End Sub

	Private Event _completed As EventHandler(Of PropertyChangedCoroutineCompletedEventArgs)
	Public Custom Event Completed As EventHandler(Of PropertyChangedCoroutineCompletedEventArgs)
	  AddHandler(ByVal value As EventHandler(Of PropertyChangedCoroutineCompletedEventArgs))
		If IsCompleted Then
		  value(Me, CompletedEventArgs)
		  Return
		End If
		AddHandler _completed, value
	  End AddHandler
	  RemoveHandler(ByVal value As EventHandler(Of PropertyChangedCoroutineCompletedEventArgs))
		  RemoveHandler _completed, value
	  End RemoveHandler
		RaiseEvent(ByVal sender As Object, ByVal e As PropertyChangedCoroutineCompletedEventArgs)
			RaiseEvent _completed(sender, e)
		End RaiseEvent
	End Event

	Protected Sub CallCompletedActions()
	  Dim actions = _notifyCompletedActions
	  _notifyCompletedActions = Nothing
	  If Nothing Is actions Then
		  Return
	  End If
	  actions(CompletedEventArgs)
	End Sub

	Private _notifyCompletedActions As Action(Of INotifyCompletedArgs)

  End Class

  Public Class PropertyChangedCoroutineCompletedEventArgs
	  Inherits EventArgs
	  Implements INotifyCompletedArgs

	''' <summary>
	''' Wait upon a <see cref="INotifyPropertyChanged.PropertyChanged"/>
	''' event for the specified property and for the property value to
	''' satisfy the <see cref="propertyValueMatchPredicate"/> test.
	''' </summary>
	''' <param name="inpc">
	''' Object that implements <see cref="INotifyPropertyChanged.PropertyChanged"/>.
	''' </param>
	''' <param name="propertyValueMatchPredicate">
	''' Function that reports true if the the operation should complete.
	''' Typically tests for a specific value of the watched property.
	''' Completes immediately if the function returns true from the start.
	''' </param>
	''' <param name="propertyName">
	''' Name of the property to watch. Null or empty means watch for any
	''' <see cref="INotifyPropertyChanged.PropertyChanged"/> event.
	''' </param>
	''' <param name="matchEmptyPropertyName">
	''' If true, will also complete if the PropertyName in the
	''' <see cref="INotifyPropertyChanged.PropertyChanged"/> EventArgs is
	''' null or empty.
	''' </param>
	''' <remarks>
	''' The <see cref="propertyValueMatchPredicate"/>, if non-null,
	''' is evaluated during construction and, if true, the operation completes
	''' immediately without bothering to wait on PropertyChanged.
	''' </remarks>
	Public Sub New(ByVal inpc As INotifyPropertyChanged, ByVal propertyValueMatchPredicate As Func(Of Boolean), ByVal propertyName As String, ByVal matchEmptyPropertyName As Boolean)

	  Me.Inpc = inpc

	  ExpectedPropertyName = propertyName
	  _matchAnyPropertyName = String.IsNullOrEmpty(ExpectedPropertyName)

	  Me.MatchEmptyPropertyName = matchEmptyPropertyName
	  _matchEmptyPropertyName = matchEmptyPropertyName OrElse String.IsNullOrEmpty(propertyName)

	  Me.PropertyValueMatchPredicate = propertyValueMatchPredicate

	  ' Done if property value already matches
	  If Nothing IsNot propertyValueMatchPredicate AndAlso propertyValueMatchPredicate() Then
		ActualPropertyName = ExpectedPropertyName
		RaiseCompleted()
	  End If

      _propertyValueMatchPredicate = If(propertyValueMatchPredicate, Function() True)
	  AddHandler inpc.PropertyChanged, AddressOf PropertyChangedHandler
	End Sub

	Private ReadOnly _matchEmptyPropertyName As Boolean
	Private ReadOnly _matchAnyPropertyName As Boolean
	Private ReadOnly _propertyValueMatchPredicate As Func(Of Boolean)

	Private Sub PropertyChangedHandler(ByVal sender As Object, ByVal e As PropertyChangedEventArgs)
	  Dim actualPropertyName = e.PropertyName
	  If _matchAnyPropertyName OrElse actualPropertyName Is ExpectedPropertyName OrElse (String.IsNullOrEmpty(actualPropertyName) AndAlso _matchEmptyPropertyName) Then

		  If Not _propertyValueMatchPredicate() Then ' didn't match required value
			  Return
		  End If

		  Me.ActualPropertyName = actualPropertyName
		  RemoveHandler Inpc.PropertyChanged, AddressOf PropertyChangedHandler ' stop listening
		  RaiseCompleted()
		  Inpc = Nothing ' let go of inpc object
	  End If
	End Sub

	Private Sub RaiseCompleted()
	  IsCompleted = True
      If Nothing Is CompletedEvent Then
        Return
      End If
	  RaiseEvent Completed(Inpc, Me)
	End Sub

	Friend Event Completed As EventHandler(Of PropertyChangedCoroutineCompletedEventArgs)
	Private privateIsCompleted As Boolean
	Public Property IsCompleted() As Boolean
		Get
			Return privateIsCompleted
		End Get
		Private Set(ByVal value As Boolean)
			privateIsCompleted = value
		End Set
	End Property
	Private privateActualPropertyName As String
	Public Property ActualPropertyName() As String
		Get
			Return privateActualPropertyName
		End Get
		Private Set(ByVal value As String)
			privateActualPropertyName = value
		End Set
	End Property
	Private privateExpectedPropertyName As String
	Public Property ExpectedPropertyName() As String
		Get
			Return privateExpectedPropertyName
		End Get
		Private Set(ByVal value As String)
			privateExpectedPropertyName = value
		End Set
	End Property
	Private privateMatchEmptyPropertyName As Boolean
	Public Property MatchEmptyPropertyName() As Boolean
		Get
			Return privateMatchEmptyPropertyName
		End Get
		Private Set(ByVal value As Boolean)
			privateMatchEmptyPropertyName = value
		End Set
	End Property
	Private privatePropertyValueMatchPredicate As Func(Of Boolean)
	Public Property PropertyValueMatchPredicate() As Func(Of Boolean)
		Get
			Return privatePropertyValueMatchPredicate
		End Get
		Private Set(ByVal value As Func(Of Boolean))
			privatePropertyValueMatchPredicate = value
		End Set
	End Property
	Private Property Inpc() As INotifyPropertyChanged


    Public ReadOnly Property Cancelled() As Boolean Implements INotifyCompletedArgs.Cancelled ' can't cancel
      Get
        Return False
      End Get
    End Property
    Public ReadOnly Property [Error]() As Exception Implements INotifyCompletedArgs.Error ' can't fail
      Get
        Return Nothing
      End Get
    End Property
    Public Property IsErrorHandled() As Boolean Implements INotifyCompletedArgs.IsErrorHandled

  End Class
End Namespace