' (c) Copyright IdeaBlade, Inc. and Microsoft Corporation
' Derived from Microsoft Silverlight User Test Framework
' This source is subject to the Microsoft Public License (Ms-PL).
' Please see http://go.microsoft.com/fwlink/?LinkID=131993 for details.
' All other rights reserved.

Imports System.ComponentModel
Imports Microsoft.Silverlight.Testing
Imports Microsoft.VisualStudio.TestTools.UnitTesting

Namespace IdeaBlade.TestFramework
  ''' <summary>
  ''' Custom test method runner class providing the ability 
  ''' to perform semi-asynchronous test tasks. 
  ''' </summary>
  ''' <remarks>
  ''' Modeled on the "WorkItemTest" class in Silverlight Unit Test Framework.
  ''' Relies upon a "Dispatcher" equivalent, the
  ''' <see cref="TestSynchronizationContext"/>
  ''' for executing queued tasks on a common thread.
  ''' </remarks>
  Public Class AsyncTestMethodRunner

	Public Sub New()
	  WorkItemContainer = New WorkItemContainer()
	  AddHandler WorkItemContainer.UnhandledException, AddressOf HandleUnhandledException

	  _syncContext = New TestSynchronizationContext()
	  AddHandler _syncContext.UnhandledException, AddressOf HandleUnhandledException
	  IsRunning = True
	End Sub

	Private ReadOnly _syncContext As TestSynchronizationContext
	Private _unhandledException As Exception
	Private ReadOnly _locker As New Object()

	Private Sub HandleUnhandledException(ByVal sender As Object, ByVal e As UnhandledExceptionEventArgs)
	  SyncLock _locker
		If Nothing IsNot _unhandledException Then ' only care about first exception
			Return
		End If
		_syncContext.ClearQueue()
		_unhandledException = If((TryCast(e.ExceptionObject, Exception)), New Exception("Unknown exception thrown in WorkItem"))
		TestComplete()
	  End SyncLock
	End Sub

	Protected Property IsRunning() As Boolean


	''' <summary>
	''' Wait for <see cref="TestComplete"/> signal that 
	''' queued async tasks are done.
	''' Typically the last line of the test method.
	''' </summary>
	''' <remarks>
	''' Perpetually processes the queue of tasks until hears the signal to stop. 
	''' Various <see cref="WaitFor"/> methods add tasks to the queue.
	''' <para>
	''' Once <see cref="WaitForTasksToFinish"/> is called, 
	''' the test method will hang until it hears the signal or the test times out.
	''' The signal is almost always the result of calling  <see cref="TestComplete"/>.
	''' </para><para>
	''' Exception catcher must raise <see cref="TestComplete"/> also
	''' to stop the loop and handle the exception.
	''' </para>
	''' </remarks>
	Protected Friend Sub WaitForTasksToFinish()

	  Do While IsRunning AndAlso WorkItemContainer.Invoke()
	  Loop

	  _syncContext.ClearQueue() ' just in case
	  ReportResult()
	End Sub

	Private Sub ReportResult()
	  Dim bugAtty = GetTestMethodBugAttribute()
	  If Nothing Is bugAtty OrElse bugAtty.Fixed Then
		If Nothing IsNot _unhandledException Then
			Throw _unhandledException
		End If
	  Else
		Dim prefix = String.Format("Test marked as a Bug: '{0}'", bugAtty.Description)
		If Nothing Is _unhandledException Then
		  Assert.Fail(prefix & "; expected test failure but didn't fail. Should mark it fixed?")
		Else
		  Console.WriteLine(prefix & " and failed as expected.")
		  Console.WriteLine(_unhandledException)
		End If
	  End If
	End Sub

	Private Function GetTestMethodBugAttribute() As BugAttribute
	  Dim callStack = New StackTrace()
	  ' Start looking for BugAttribute on test method
	  ' starting 4 frames up (count from 0)
	  For i = 3 To callStack.FrameCount - 1
		Dim frame = callStack.GetFrame(i)
		Dim method = frame.GetMethod()
		Dim bugAtty = method.GetCustomAttributes(GetType(BugAttribute), False).FirstOrDefault()
		If Nothing IsNot bugAtty Then
			Return CType(bugAtty, BugAttribute)
		End If
	  Next i
	  Return Nothing
	End Function

	''' <summary>
	''' Gets the current test task container.
	''' </summary>
	''' <returns>The current container for the test's tasks.</returns>
	Private privateWorkItemContainer As CompositeWorkItem
	<EditorBrowsable(EditorBrowsableState.Never)>
	Protected Property WorkItemContainer() As CompositeWorkItem
		Get
			Return privateWorkItemContainer
		End Get
		Private Set(ByVal value As CompositeWorkItem)
			privateWorkItemContainer = value
		End Set
	End Property

	''' <summary>
	''' Signal that a test is complete when using Async testing.  
	''' 
	''' WARNING: If you use your own methods for completing, such as an 
	''' HtmlTimer or other threading method, it is possible that this call 
	''' will occur *AFTER* the test has timed out when using Timeouts.  As 
	''' such, be very careful as you could complete the call to the *next*  test.
	''' </summary>
	Public Overridable Sub TestComplete()
	  SyncLock _locker
		_syncContext.ClearQueue()
		WorkItemContainer.WorkItemCompleteInternal()
	  End SyncLock
	End Sub

	''' <summary>
	''' Get thread id of <see cref="System.Threading.SynchronizationContext"/>
	''' used by this test.
	''' </summary>
	Public ReadOnly Property SyncContextThreadId() As Integer
		Get
			Return _syncContext.SyncContextThreadId
		End Get
	End Property

	''' <summary>
	''' Add a task object to the test queue.  For a test that is currently 
	''' executing, all tasks contained within the queue are executed to 
	''' completion (unless an Exception is thrown) -before- moving on to 
	''' the next test.
	''' </summary>
	''' <param name="testTaskObject">Asynchronous test task instance.</param>
	Public Overridable Sub EnqueueWorkItem(ByVal testTaskObject As WorkItem)
		WorkItemContainer.Enqueue(testTaskObject)
	End Sub

	''' <summary>
	''' Delay a minimum amount of time before continuing. Similar to a sleep
	''' call that is not a blocking call.
	''' </summary>
	''' <param name="delay">The minimum time span to wait before continuing.</param>
	Public Overridable Sub EnqueueDelay(ByVal delay As TimeSpan)
	  EnqueueWorkItem(CreateDelay(delay))
	End Sub

	''' <summary>
	''' Creates a delay work item instance.
	''' </summary>
	''' <param name="delay">The minimum time span to wait before continuing.</param>
	''' <returns>Returns a new work item.</returns>
	<EditorBrowsable(EditorBrowsableState.Never)>
	Public Function CreateDelay(ByVal delay As TimeSpan) As WorkItem
	  Return New SleepWorkItem(delay)
	End Function

	''' <summary>
	''' Delay a minimum amount of time before continuing. Similar to a sleep
	''' call that is not a blocking call.
	''' </summary>
	''' <param name="milliseconds">The minimum number of milliseconds to wait
	''' until the delay is finished.</param>
	Public Overridable Sub EnqueueDelay(ByVal milliseconds As Double)
	  EnqueueDelay(TimeSpan.FromMilliseconds(milliseconds))
	End Sub

	''' <summary>
	''' Requires a bool returning delegate to be passed in. Instructs the 
	''' test task queue to wait until the conditional call returns True to 
	''' continue executing other test tasks and/or ending the test method.
	''' </summary>
	''' <param name="conditionalDelegate">Conditional method or delegate. 
	''' Test will halt until this condition returns True.</param>
	Public Overridable Sub EnqueueConditional(ByVal conditionalDelegate As Func(Of Boolean))
	  EnqueueWorkItem(CreateConditional(conditionalDelegate))
	End Sub

	''' <summary>
	''' Creates a conditional work item instance.
	''' </summary>
	''' <param name="conditionalDelegate">Conditional method or delegate. 
	''' Test will halt until this condition returns True.</param>
	''' <returns>Returns a new work item.</returns>
	<EditorBrowsable(EditorBrowsableState.Never)>
	Public Function CreateConditional(ByVal conditionalDelegate As Func(Of Boolean)) As WorkItem
	  Return New ConditionalWorkItem(conditionalDelegate, _syncContext)
	End Function

	''' <summary>
	''' Enqueue a test task which calls the TestComplete method of 
	''' SilverlightTest.
	''' </summary>
	Public Overridable Sub EnqueueTestComplete()
	  ' Don't do with SyncContext
	  EnqueueWorkItem(New CallbackWorkItem(AddressOf TestComplete))
	  'EnqueueCallback(TestComplete); // was
	End Sub

	''' <summary>
	''' Add a Callback method into the test task queue.  Similar to the 
	''' PumpMessages(...) call, with the difference being that there is no 
	''' longer a single requirement: you can enqueue several callback 
	''' methods and other test tasks, all of which will execute before the 
	''' test completes and/or the engine continues.
	''' </summary>
	''' <param name="testCallbackDelegate">Void-returning delegate, 
	''' anonymous delegates work fine too.</param>
	Public Overridable Sub EnqueueCallback(ByVal testCallbackDelegate As Action)
	  EnqueueWorkItem(New CallbackWorkItem(testCallbackDelegate, _syncContext))
	End Sub

	' WB: No usage found 
	'/// <summary>
	'/// Creates a callback work item instance.
	'/// </summary>
	'/// <param name="testCallbackDelegate">Void-returning delegate, 
	'/// anonymous delegates work fine too.</param>
	'/// <returns>Returns a new work item.</returns>
	'[EditorBrowsable(EditorBrowsableState.Never)]
	'public WorkItem CreateCallback(Action testCallbackDelegate) {
	'  return new CallbackWorkItem(testCallbackDelegate);
	'}

	''' <summary>
	''' Adds a number of callback methods into the test task queue.
	''' </summary>
	''' <param name="actions">Set of Action instances.</param>
	Public Sub EnqueueCallback(ParamArray ByVal actions() As Action)
	  If actions IsNot Nothing Then
		For Each action As Action In actions
		  EnqueueCallback(action)
		Next action
	  End If
	End Sub

	''' <summary>
	''' Process an exception using the test engine logic for 
	''' ExpectedExceptions and logging the exception if needed.
	''' </summary>
	''' <param name="ex">Exception object.</param>
	Public Overridable Sub HandleException(ByVal ex As Exception)
	  WorkItemContainer.WorkItemException(ex)
	End Sub

  End Class

  Friend Class WorkItemContainer
	  Inherits CompositeWorkItem
	Public Sub New()
	  FinishWhenEmpty = False
	End Sub
  End Class
End Namespace

' Placeholder namespace to prevent compile failures
' when test files are linked to both Desktop and Silverlight projects.
' ReSharper disable EmptyNamespace
Namespace Microsoft.Silverlight.Testing
End Namespace
' ReSharper restore EmptyNamespace