' (c) Copyright IdeaBlade, Inc. and Microsoft Corporation
' Derived from Microsoft Silverlight User Test Framework
' This source is subject to the Microsoft Public License (Ms-PL).
' Please see http://go.microsoft.com/fwlink/?LinkID=131993 for details.
' All other rights reserved.

Imports System.Diagnostics.CodeAnalysis

Namespace IdeaBlade.TestFramework
  ''' <summary>
  ''' TestWorkItem which can contain sub-tasks; the underlying work item is 
  ''' not marked complete until the Children have completed or an Exception 
  ''' is thrown.
  ''' </summary>
  Public Class CompositeWorkItem
	  Inherits WorkItem
	''' <summary>
	''' Store the underlying tasks.
	''' </summary>
	Private ReadOnly _workItemQueue As Queue(Of WorkItem)

	''' <summary>
	''' Whether the TestTaskContainer::IsComplete should be set to true 
	''' after all children have been de-queued.
	''' </summary>
	Private _finishWhenEmpty As Boolean

	''' <summary>
	''' Whether the first invoke has happened yet or not.
	''' </summary>
	Private _invoked As Boolean

	''' <summary>
	''' Constructor for the TestTaskContainer type.
	''' </summary>
	Public Sub New()
	  _workItemQueue = New Queue(Of WorkItem)()
	  _finishWhenEmpty = True
	End Sub

	''' <summary>
	''' Gets a value indicating whether the container is marked Complete 
	''' when all children have executed or not.
	''' </summary>
	Public Property FinishWhenEmpty() As Boolean
	  Get
		  Return _finishWhenEmpty
	  End Get

	  Protected Set(ByVal value As Boolean)
		  _finishWhenEmpty = value
	  End Set
	End Property

	''' <summary>
	''' Optional method to call on the first invoke.
	''' </summary>
	Protected Overridable Sub FirstInvoke()
	End Sub

	''' <summary>
	''' Invoke the test container; in turn will execute child work items 
	''' as needed. Supports executing multiple items immediately for
	''' performance reasons.
	''' </summary>
	''' <returns>True if additional work remains, False once IsComplete 
	''' is set to true.</returns>
	Public Overrides Function Invoke() As Boolean ' return false when complete + event
	  Dim workItem As WorkItem
	  Dim additionalWork As Boolean
	  Dim canContinueImmediateExecution As Boolean = False

	  Do
		additionalWork = Invoke(workItem)
		If workItem IsNot Nothing Then
		  canContinueImmediateExecution = workItem.CanExecuteImmediately
		End If
	  Loop While additionalWork AndAlso canContinueImmediateExecution

	  Return additionalWork
	End Function

	''' <summary>
	''' Invoke the test container; in turn will execute child work items 
	''' as needed.
	''' </summary>
	''' <param name="usedWorkItem">The work item used for the invoke.</param>
	''' <returns>True if additional work remains, False once IsComplete 
	''' is set to true.</returns>
    <SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification:="Key part of the unit test engine")>
    Private Overloads Function Invoke(ByRef usedWorkItem As WorkItem) As Boolean
      usedWorkItem = Nothing

      ' On the first invoke, run the optional method
      If Not _invoked Then
        _invoked = True
        FirstInvoke()
      End If

      ' If this container is complete
      If IsComplete Then
        ClearChildren()
        Return False
      End If

      ' If no remaining tasks
      If (Not RemainingWork) AndAlso _finishWhenEmpty Then
        ClearChildren()
        WorkItemComplete()
        Return False
      End If

      ' The completion comes from an external event, background task, or 
      ' other delegate outside of the empty complete flag
      If Not RemainingWork Then
        Return True
      End If

      ' Examine the status of child work items
      usedWorkItem = Peek()

      ' Assumption: The task is running by default
      Dim taskStillRunning As Boolean = True

      Try
        taskStillRunning = usedWorkItem.Invoke()
      Catch exp As Exception
        taskStillRunning = False
        OnUnhandledException(exp)
      Finally
        If Not taskStillRunning Then
          If RemainingWork Then
            ' Remove the child
            Dequeue()
          Else
            WorkItemComplete()
          End If
        End If
      End Try
      Return True
    End Function

	''' <summary>
	''' Gets a value indicating whether sub tasks remain.
	''' </summary>
	Public ReadOnly Property RemainingWork() As Boolean
	  Get
		SyncLock _workItemQueue
		  Return _workItemQueue.Count > 0
		End SyncLock
	  End Get
	End Property

	''' <summary>
	''' Clear the children.
	''' </summary>
	Protected Sub ClearChildren()
	  SyncLock _workItemQueue
		_workItemQueue.Clear()
	  End SyncLock
	End Sub

	''' <summary>
	''' Dequeue a work item.
	''' </summary>
	''' <returns>A work item.</returns>
	Public Function Dequeue() As WorkItem
	  SyncLock _workItemQueue
		Return _workItemQueue.Dequeue()
	  End SyncLock
	End Function

	''' <summary>
	''' Add a new work item to the container to schedule it for invocation.
	''' </summary>
	''' <param name="item">New test work item to enqueue.</param>
	Public Sub Enqueue(ByVal item As WorkItem)
	  SyncLock _workItemQueue
		_workItemQueue.Enqueue(item)
	  End SyncLock
	End Sub

	''' <summary>
	''' Adds a new work item that will execute more quickly.
	''' </summary>
	''' <param name="item">The item of work.</param>
	Public Sub EnqueueQuick(ByVal item As WorkItem)
	  item.CanExecuteImmediately = True
	  Enqueue(item)
	End Sub

	''' <summary>
	''' Enqueues a callback or action that will quickly execute.
	''' </summary>
	''' <param name="action">The action or method.</param>
	Public Sub EnqueueQuick(ByVal action As Action)
	  EnqueueQuick(New CallbackWorkItem(action))
	End Sub

	''' <summary>
	''' Add a new callback action and schedule it for invocation.
	''' </summary>
	''' <param name="action">The action.</param>
	Public Sub Enqueue(ByVal action As Action)
	  Enqueue(New CallbackWorkItem(action))
	End Sub

	''' <summary>
	''' Return the top work item, if any, from this container.
	''' </summary>
	''' <returns>Peek into any test work item.</returns>
	Public Function Peek() As WorkItem
	  SyncLock _workItemQueue
		Return _workItemQueue.Peek()
	  End SyncLock
	End Function

	''' <summary>
	''' Work items must call this method to indicate completion of the work 
	''' item; in turn fires the Complete event delegates.
	''' </summary>
	Protected Overrides Sub WorkItemComplete()
	  MyBase.WorkItemComplete()
	  OnComplete(EventArgs.Empty)
	End Sub

	''' <summary>
	''' Internal-only version which can be called during a test completion 
	''' through the relation - not necessarily the best design; events 
	''' may make more sense long-term.
	''' </summary>
	Friend Overridable Sub WorkItemCompleteInternal()
	  WorkItemComplete()
	End Sub

	''' <summary>
	''' Call when an exception occurs inside a work item.
	''' </summary>
	''' <param name="e">Exception object.</param>
	Friend Overridable Sub WorkItemException(ByVal e As Exception)
	  ' REVIEW: DESIGN REVIEW OF EXCEPTION HANDLING NEEDED
	  OnUnhandledException(e)
	  WorkItemComplete()
	End Sub

	''' <summary>
	''' Complete event is fired when the underlying WorkItemComplete method 
	''' is called by the work item.
	''' </summary>
	Public Event Complete As EventHandler

	''' <summary>
	''' Event fired when an exception is thrown and unhandled within the 
	''' underlying Invoke sequence.
	''' </summary>
	Public Event UnhandledException As EventHandler(Of UnhandledExceptionEventArgs)

	''' <summary>
	''' Fire the Complete event.
	''' </summary>
	''' <param name="e">Empty event arguments.</param>
	Protected Overridable Sub OnComplete(ByVal e As EventArgs)
      If e IsNot Nothing Then
        RaiseEvent Complete(Me, e)
      End If
	End Sub

	''' <summary>
	''' Fire the unhandled exception event.
	''' </summary>
	''' <param name="exception">Exception object.</param>
	Protected Overridable Sub OnUnhandledException(ByVal exception As Exception)
	  ' REVIEW: Should this method take the actual Exception, or the actual EventArgs for containing it?
      If exception IsNot Nothing Then
        RaiseEvent UnhandledException(Me, New UnhandledExceptionEventArgs(exception, False))
      End If
	End Sub

	''' <summary>
	''' Gets or sets a value indicating whether the task can execute
	''' immediately.
	''' </summary>
	Public Overrides Property CanExecuteImmediately() As Boolean
	  Get
		Return MyBase.CanExecuteImmediately
	  End Get
	  Set(ByVal value As Boolean)
		' Silently ignore this operation as it could cause unreliable
		' results.
	  End Set
	End Property
  End Class
End Namespace