Imports System.Threading

Namespace IdeaBlade.TestFramework

  ''' <summary>
  ''' Derived WorkItems may choose to ask a <see cref="SynchronizationContext"/>
  ''' to execute their actions.
  ''' See <see cref="SyncContext"/>.
  ''' </summary>
  Public MustInherit Class SynchronizedWorkItem
	  Inherits WorkItem

	#Region "SynchronizationContext related"

	''' <summary>
	''' Get and set <see cref="SynchronizationContext"/> to be used
	''' for <see cref="Send(Action)"/> and <see cref="Post(Action)"/> calls.
	''' </summary>
	''' <remarks>
	''' Get guarantees that such a SyncContext exists; if it can't find one
	''' it creates one ... which simply executes the action on the current thread.
	''' </remarks>
	Protected Property SyncContext() As SynchronizationContext
	  Get
		If _synchronizationContext IsNot Nothing Then
			Return _synchronizationContext
		Else
			_synchronizationContext = If(SynchronizationContext.Current, New SynchronizationContext())
			Return _synchronizationContext
		End If
	  End Get
	  Set(ByVal value As SynchronizationContext)
		  _synchronizationContext = value
	  End Set
	End Property

	Private _synchronizationContext As SynchronizationContext

	''' <summary>
	''' Send the action to the <see cref="SyncContext"/> for execution.
	''' See also <see cref="SynchronizationContext.Send"/>.
	''' </summary>
	''' <remarks>Uses a thread-local counter for the state object.</remarks>
	Protected Sub Send(ByVal act As Action)
	  Send(act, _stateCounter)
	  _stateCounter += 1
	End Sub

	''' <summary>
	''' Send the action and state object to the <see cref="SyncContext"/> for execution.
	''' See also <see cref="SynchronizationContext.Send"/>.
	''' </summary>
	Protected Sub Send(ByVal act As Action, ByVal state As Object)
      Send(Sub(s) act(), state)
	End Sub

	''' <summary>
	''' Send the action and state object to the <see cref="SyncContext"/> for execution.
	''' See also <see cref="SynchronizationContext.Send"/>.
	''' </summary>
	Protected Sub Send(ByVal act As SendOrPostCallback, ByVal state As Object)
	  SyncContext.Send(act, state)
	End Sub

	''' <summary>
	''' Post the action to the <see cref="SyncContext"/> for execution.
	''' See also <see cref="SynchronizationContext.Post"/>.
	''' </summary>
	''' <remarks>Uses a thread-local counter for the state object.</remarks>
	Protected Sub Post(ByVal act As Action)
	  Post(act, _stateCounter)
	  _stateCounter += 1
	End Sub

	''' <summary>
	''' Post the action and state object to the <see cref="SyncContext"/> for execution.
	''' See also <see cref="SynchronizationContext.Post"/>.
	''' </summary>
	Protected Sub Post(ByVal act As Action, ByVal state As Object)
      Post(Sub(s) act(), state)
	End Sub
	''' <summary>
	''' Post the action and state object to the <see cref="SyncContext"/> for execution.
	''' See also <see cref="SynchronizationContext.Post"/>.
	''' </summary>
	Protected Sub Post(ByVal act As SendOrPostCallback, ByVal state As Object)
	  SyncContext.Post(act, state)
	End Sub

	<ThreadStatic>
	Private Shared _stateCounter As Integer

	#End Region

  End Class
End Namespace