Imports System.Security.Permissions
Imports System.Threading

' ReSharper disable RedundantNameQualifier

Namespace IdeaBlade.TestFramework

  ''' <summary>
  ''' A Faked <see cref="SynchronizationContext"/> for Desktop Async Tests.
  ''' </summary>
  ''' <remarks>
  ''' This synchronization context implementation comes from Brian Noyes'
  ''' Prism-for-Windows-Forms implementation
  ''' and is based on the ThreadPoolSynchronizer sample
  ''' from IDesign Inc, www.idesign.net.
  ''' Brian emailed it to Ward on 15 Aug 2010.
  ''' <para>
  ''' Added handling of unhandled exceptions thrown in WorkItem.
  ''' Changed name from SynchronizationContextMock
  ''' and namespace from IDesign.Composite.Testing.Mocks.
  ''' Changed "WorkItem" class to "SyncWorkItem" because conflicted
  ''' with my "WorkItem" class in this namespace.
  ''' "Cleaned-up" to mostly "conform" to IdeaBlade coding standards.
  ''' Added some sugar (extra ctor, simplifying overloads of Send/Post).
  ''' </para>
  ''' </remarks> 
  <SecurityPermission(SecurityAction.Demand, ControlThread := True)>
  Public Class TestSynchronizationContext
	  Inherits SynchronizationContext
	  Implements IDisposable

	Public Const TestSynchronizationContextTheadName As String = "TestSynchronizationContext"

	Private ReadOnly _workerThread As WorkerThread
	Private ReadOnly _syncWorkItemQueue As Queue(Of SyncWorkItem)

	<ThreadStatic>
	Private _userStateCounter As Integer

	Public Sub New()
		Me.New(TestSynchronizationContextTheadName)
	End Sub

	Public Sub New(ByVal contextName As String)
	  CallQueued = New Semaphore(0, Int32.MaxValue)
	  _syncWorkItemQueue = New Queue(Of SyncWorkItem)()
	  _workerThread = New WorkerThread(contextName, Me)
	End Sub

	Public ReadOnly Property SyncContextThreadId() As Integer
	  Get
		  Return _workerThread.ThreadObj.ManagedThreadId
	  End Get
	End Property

	Private privateCallQueued As Semaphore
	Protected Property CallQueued() As Semaphore
		Get
			Return privateCallQueued
		End Get
		Private Set(ByVal value As Semaphore)
			privateCallQueued = value
		End Set
	End Property

	Friend Overridable Sub QueueWorkItem(ByVal workItem As SyncWorkItem)
	  SyncLock _syncWorkItemQueue
		_syncWorkItemQueue.Enqueue(workItem)
		CallQueued.Release()
	  End SyncLock
	End Sub

	Protected Overridable ReadOnly Property QueueEmpty() As Boolean
	  Get
		SyncLock _syncWorkItemQueue
		  Return _syncWorkItemQueue.Count <= 0
		End SyncLock
	  End Get
	End Property

	Friend Overridable Function GetNext() As SyncWorkItem
	  CallQueued.WaitOne()
	  SyncLock _syncWorkItemQueue
		Return If(_syncWorkItemQueue.Count = 0, Nothing, _syncWorkItemQueue.Dequeue())
	  End SyncLock
	End Function


	Public Overrides Function CreateCopy() As SynchronizationContext
		Return Me
	End Function

	Public Overloads Sub Post(ByVal act As Action)
      Post(Sub(state) act(), _userStateCounter)
	  _userStateCounter += 1
	End Sub

	Public Overrides Sub Post(ByVal method As SendOrPostCallback, ByVal state As Object)
	  Dim workItem = New SyncWorkItem(method, state)
	  QueueWorkItem(workItem)
	End Sub

	Public Overloads Sub Send(ByVal act As Action)
      Send(Sub(state) act(), _userStateCounter)
	  _userStateCounter += 1
	End Sub

	Public Overrides Sub Send(ByVal method As SendOrPostCallback, ByVal state As Object)
	  If SynchronizationContext.Current Is Me Then
		'If already on the correct context, must invoke now to avoid deadlock
		ExecuteNow(method, state)
		' method(state); //this original line wouldn't manage exceptions consistently
		Return
	  End If
	  Dim workItem = New SyncWorkItem(method, state)
	  QueueWorkItem(workItem)
	  workItem.Wait()
	End Sub

	Private Sub ExecuteNow(ByVal method As SendOrPostCallback, ByVal state As Object)
	  Try
		method(state)
	  Catch ex As Exception
		_workerThread.RaiseUnhandledException(ex)
	  End Try
	End Sub

	Public Sub ClearQueue()
	  SyncLock _syncWorkItemQueue
		_syncWorkItemQueue.Clear()
	  End SyncLock
	End Sub

	Public Custom Event UnhandledException As EventHandler(Of UnhandledExceptionEventArgs)
	  AddHandler(ByVal value As EventHandler(Of UnhandledExceptionEventArgs))
		  AddHandler _workerThread.UnhandledException, value
      End AddHandler

	  RemoveHandler(ByVal value As EventHandler(Of UnhandledExceptionEventArgs))
		  RemoveHandler _workerThread.UnhandledException, value
      End RemoveHandler

      RaiseEvent(ByVal sender As System.Object, ByVal e As UnhandledExceptionEventArgs)
        _workerThread.RaiseUnhandledException(CType(e.ExceptionObject, Exception))  'DEK HACK
        'RaiseEvent _workerThread.UnhandledException(sender, e)  'Translation offered by Instant VB
      End RaiseEvent

	End Event

	Public Sub Dispose() Implements IDisposable.Dispose
	  Close()
	End Sub

	Public Sub Close()
	  If CallQueued.SafeWaitHandle.IsClosed Then
		Return
	  End If
	  CallQueued.Release(Int32.MaxValue)
	  _workerThread.Kill()
	  _workerThread.ClearUnhandledExceptionHandlers()
	  CallQueued.Close()
	End Sub

	Public Sub Abort()
	  CallQueued.Release(Int32.MaxValue)
	  _workerThread.ThreadObj.Abort()
	  _workerThread.ClearUnhandledExceptionHandlers()
	  CallQueued.Close()
	End Sub

	#Region "WorkThread"

	Friend Class WorkerThread

	  Public ReadOnly ThreadObj As Thread
	  Private ReadOnly _syncContext As TestSynchronizationContext
	  Private _endLoop As Boolean

	  Friend Sub New(ByVal name As String, ByVal context As TestSynchronizationContext)
		_syncContext = context

		_endLoop = False

		ThreadObj = New Thread(AddressOf Run) With {.IsBackground = True, .Name = name}
		ThreadObj.Start()
	  End Sub

	  Friend Property EndLoop() As Boolean
		Set(ByVal value As Boolean)
			SyncLock Me
				_endLoop = value
			End SyncLock
		End Set
		Get
			SyncLock Me
				Return _endLoop
			End SyncLock
		End Get
	  End Property

	  Friend Sub Run()
		SetSynchronizationContext(_syncContext)

		Do While EndLoop = False
		  ExecuteWorkItem(_syncContext.GetNext())
		Loop
	  End Sub

	  Friend Sub ExecuteWorkItem(ByVal workItem As SyncWorkItem)
		If Nothing Is workItem Then
			Return
		End If
		Dim ex As Exception = workItem.CallBack()
		If Nothing IsNot ex Then
			RaiseUnhandledException(ex)
		End If
	  End Sub

	  #Region "Exception handling"

	  Friend Custom Event UnhandledException As EventHandler(Of UnhandledExceptionEventArgs)
		AddHandler(ByVal value As EventHandler(Of UnhandledExceptionEventArgs))
			SyncLock Me
				AddHandler _unhandledException, value
			End SyncLock
		End AddHandler
		RemoveHandler(ByVal value As EventHandler(Of UnhandledExceptionEventArgs))
			SyncLock Me
				RemoveHandler _unhandledException, value
			End SyncLock
		End RemoveHandler
		  RaiseEvent(ByVal sender As Object, ByVal e As UnhandledExceptionEventArgs)
			  RaiseEvent _unhandledException(sender, e)
		  End RaiseEvent
	  End Event

	  ' ReSharper disable InconsistentNaming
	  Private Event _unhandledException As EventHandler(Of UnhandledExceptionEventArgs)
	  ' ReSharper restore InconsistentNaming

	  Friend Sub RaiseUnhandledException(ByVal ex As Exception)
		OnUnhandledException(New UnhandledExceptionEventArgs(ex, False))
	  End Sub

	  Protected Overridable Sub OnUnhandledException(ByVal e As UnhandledExceptionEventArgs)
        SyncLock Me
          RaiseEvent _unhandledException(Me, e)
        End SyncLock
        If Nothing Is e Then
          Return
        End If
        RaiseUnhandledException(CType(e.ExceptionObject, Exception))
      End Sub
	  Friend Sub ClearUnhandledExceptionHandlers()
        SyncLock Me
          RemoveHandler _unhandledException, Nothing
        End SyncLock
	  End Sub

	  #End Region

	  Public Sub Kill()
		'Kill is called on client thread - must use cached thread object
		Debug.Assert(ThreadObj IsNot Nothing)
		If ThreadObj.IsAlive = False Then
		  Return
		End If
		EndLoop = True

		'Wait for thread to die
		ThreadObj.Join()
	  End Sub
	End Class

	#End Region

  End Class

  <Serializable>
  Friend Class SyncWorkItem

	Private ReadOnly _userState As Object
	Private ReadOnly _method As SendOrPostCallback
	Private ReadOnly _asyncWaitHandle As ManualResetEventSlim

	Friend Sub Wait()
	  _asyncWaitHandle.Wait()
	End Sub

	Friend Sub New(ByVal method As SendOrPostCallback, ByVal state As Object)
	  _method = method
	  _userState = state
	  _asyncWaitHandle = New ManualResetEventSlim(False)
	End Sub

	' Don't execute; just terminate
	Friend Sub Skip()
	  _asyncWaitHandle.Set()
	End Sub

	'This method is called on the worker thread to execute the method
	Friend Function CallBack() As Exception
	  Dim except As Exception = Nothing
	  Try
		_method(_userState)
	  Catch ex As Exception
		except = ex
	  End Try
	  _asyncWaitHandle.Set()
	  Return except
	End Function
  End Class


End Namespace
