Imports System.Reflection
Imports IdeaBlade.EntityModel

Namespace IdeaBlade.TestFramework

  Partial Public Class DevForceTest

	''' <summary>
	''' Try synchronous act and then complete (end) the test.
	''' Catch exception and route it to test harness.
	''' </summary>
	Protected Sub TryToComplete(ByVal act As Action)
	  [Try](act)
	  WaitFor.TestComplete()
	End Sub

	''' <summary>
	''' Try synchronous act.
	''' Catch exception and route it to test harness.</summary>
	Protected Sub [Try](ByVal act As Action)
	  WaitFor.It(act)
	End Sub

	Protected ReadOnly WaitFor As NestedWaitFor

	''' <summary>
	''' An attempt at more intuitive synonyms for 
	''' Silverlight Testing Framework's Enqueue methods.
	''' Also affords consistency with WaitFor extension methods,
	''' <see cref="OperationTestForErrorExtensions"/>.
	''' </summary>
	''' <remarks>
	''' A coding experiment. Not sure this is such a great idea
	''' Perhaps more clever than good.
	''' </remarks>
	Protected Class NestedWaitFor

	  Public Sub New(ByVal devForceTest As DevForceTest)
		DevForceTest = devForceTest
	  End Sub

	  Private privateDevForceTest As DevForceTest
	  Protected Property DevForceTest() As DevForceTest
		  Get
			  Return privateDevForceTest
		  End Get
		  Private Set(ByVal value As DevForceTest)
			  privateDevForceTest = value
		  End Set
	  End Property

	  ''' <summary>
	  ''' Add a Callback method into the test task queue. 
	  ''' Synonym for the <see cref="Callback"/> method.
	  ''' </summary>
	  ''' <param name="testCallback">
	  ''' Void-returning delegate, anonymous delegates work fine too.
	  ''' </param>
	  Public Sub It(ByVal testCallback As Action)
		DevForceTest.WaitForIt(testCallback)
	  End Sub

	  ''' <summary>
	  ''' Add a Callback method into the test task queue. 
	  ''' Similar to the PumpMessages(...) call, with the difference being 
	  ''' that there is no longer a single requirement: 
	  ''' you can enqueue several callback methods and other test tasks, 
	  ''' all of which will execute before the test completes and/or the engine continues.
	  ''' </summary>
	  ''' <param name="testCallback">
	  ''' Void-returning delegate, anonymous delegates work fine too.
	  ''' </param>
	  Public Sub Callback(ByVal testCallback As Action)
		DevForceTest.EnqueueCallback(testCallback)
	  End Sub

	  Public Sub Delay(ByVal milliseconds As Double)
		DevForceTest.EnqueueDelay(milliseconds)
	  End Sub
	  ''' <summary>
	  ''' Delay a minimum amount of time before continuing. Similar to a sleep call that is not a blocking call.
	  ''' </summary>
	  ''' <param name="delay">
	  ''' The minimum time span to wait before continuing
	  ''' </param>
	  Public Sub Delay(ByVal delay As TimeSpan)
		DevForceTest.EnqueueDelay(delay)
	  End Sub
	  ' Based on Justin Angel: http://blogs.silverlight.net/blogs/justinangel/archive/2009/02/25/silverlight-unit-testing-rhinomocks-unity-and-resharper.aspx
	  Public Function [Event](Of T)(ByVal objectToWaitForItsEvent As T, ByVal eventName As String) As T

		Dim eventInfo As EventInfo = objectToWaitForItsEvent.GetType().GetEvent(eventName)

		Dim eventRaised As Boolean = False

		If GetType(RoutedEventHandler).IsAssignableFrom(eventInfo.EventHandlerType) Then
		  eventInfo.AddEventHandler(objectToWaitForItsEvent, CType(Sub() eventRaised = True, RoutedEventHandler))

		ElseIf GetType(EventHandler).IsAssignableFrom(eventInfo.EventHandlerType) Then
		  eventInfo.AddEventHandler(objectToWaitForItsEvent, CType(Sub() eventRaised = True, EventHandler))
		End If

		DevForceTest.EnqueueConditional(Function() eventRaised)

		Return objectToWaitForItsEvent
	  End Function

	  ''' <summary>
	  ''' Wait until the DevForce <see cref="BaseOperation.IsCompleted"/> flag is true.
	  ''' Uses the <see cref="Conditional"/> method.
	  ''' </summary>
	  ''' <param name="op">Operation object returned from a DevForce async method.</param>
	  Public Sub IsCompleted(ByVal op As BaseOperation)
		IsCompleted(Function() op)
	  End Sub

	  ''' <summary>
	  ''' Wait until the DevForce <see cref="BaseOperation.IsCompleted"/> flag is true.
	  ''' Uses the <see cref="Conditional"/> method.
	  ''' </summary>
	  ''' <param name="opFn">Fnc rtning operation object returned from a DevForce async method.</param>
	  Public Sub IsCompleted(ByVal opFn As Func(Of BaseOperation))
		DevForceTest.EnqueueConditional(Function()
			Dim op = opFn()
			Return op.IsCompleted OrElse op.HasError OrElse op.Cancelled
		End Function)
	  End Sub

	  ''' <summary>
	  ''' Instructs the test task queue to wait until the conditional call returns True 
	  ''' to continue executing other test tasks and/or ending the test method.
	  ''' </summary>
	  ''' <param name="conditional">
	  ''' Conditional method or delegate. Test will halt until this condition returns True.
	  ''' </param>
	  Public Sub Conditional(ByVal conditional As Func(Of Boolean))
		DevForceTest.EnqueueConditional(conditional)
	  End Sub

	  ''' <summary>
	  ''' Enqueue a test task which calls the <see cref="TestComplete"/> method.
	  ''' </summary>
	  Public Sub TestComplete()
		DevForceTest.EnqueueTestComplete()
	  End Sub
	  ''' <summary>
	  ''' Wait for <see cref="TestComplete"/> signal that 
	  ''' queued async tasks are done.
	  ''' See <see cref="WaitForTasksToFinish"/>.
	  ''' </summary>
	  Public Sub TasksToFinish()
		DevForceTest.WaitForTasksToFinish()
	  End Sub
	End Class
  End Class
End Namespace