Imports System.Collections
Imports IdeaBlade.EntityModel
Imports Microsoft.VisualStudio.TestTools.UnitTesting

Namespace IdeaBlade.TestFramework

  Public Module OperationTestTheResultsExtensions

	''' <summary>
	''' Test the results of a successful <see cref="EntityQuery"/>
	''' for conformance with expectations.
	''' Raises TestComplete after tests.
	''' Throws <see cref="Assert.Fail(string)"/> if error or cancelled
	''' using <see cref="queryDescription"/> in the message.
	''' </summary>
	<System.Runtime.CompilerServices.Extension> _
	Public Function TestTheResults(ByVal op As EntityQueryOperation, ByVal action As Action(Of IEnumerable), Optional ByVal queryDescription As String = Nothing) As EntityQueryOperation

	  If String.IsNullOrWhiteSpace(queryDescription) Then
		queryDescription = "EntityQueryOperation"
	  End If

	  AddHandler op.Completed, Sub(s, args)
		ReportFailures(args, queryDescription)
		If Nothing IsNot action Then
			action(args.Results)
		End If
		WaitFor.TestComplete()
	  End Sub

	  Return op

	End Function

	''' <summary>
	''' Test the results of a successful <see cref="EntityQuery{T}"/>
	''' for conformance with expectations.
	''' Raises TestComplete after tests.
	''' Throws <see cref="Assert.Fail(string)"/> if error or cancelled
	''' using <see cref="queryDescription"/> in the message.
	''' </summary>
	<System.Runtime.CompilerServices.Extension> _
	Public Function TestTheResults(Of T)(ByVal op As EntityQueryOperation(Of T), ByVal action As Action(Of IEnumerable(Of T)), Optional ByVal queryDescription As String = Nothing) As EntityQueryOperation(Of T)

	  If String.IsNullOrWhiteSpace(queryDescription) Then
		queryDescription = String.Format("EntityQueryOperation<{0}", GetType(T).Name)
	  End If

	  AddHandler op.Completed, Sub(s, args)
		ReportFailures(args, queryDescription)
		If Nothing IsNot action Then
			action(args.Results)
		End If
		WaitFor.TestComplete()
	  End Sub

	  Return op

	End Function


	''' <summary>
	''' Test the result of a successful <see cref="EntityScalarQuery{T}"/>
	''' for conformance with expectations.
	''' Raises TestComplete after tests.
	''' Throws <see cref="Assert.Fail(string)"/> if error or cancelled
	''' using <see cref="queryDescription"/> in the message.
	''' </summary>
	<System.Runtime.CompilerServices.Extension> _
	Public Function TestTheResult(Of T)(ByVal op As EntityScalarQueryOperation(Of T), ByVal action As Action(Of T), Optional ByVal queryDescription As String = Nothing) As EntityScalarQueryOperation(Of T)

		If String.IsNullOrWhiteSpace(queryDescription) Then
		queryDescription = String.Format("EntityScalarQueryOperation<{0}", GetType(T).Name)
		End If

	  AddHandler op.Completed, Sub(s, args)
		ReportFailures(args, queryDescription)
		If Nothing IsNot action Then
			action(args.Result)
		End If
		WaitFor.TestComplete()
	  End Sub

	  Return op

	End Function

	''' <summary>
	''' Test the result of a successful <see cref="EntityQuery"/>
	''' from a <see cref="EntityQueryCoroutineOperation"/>
	''' for conformance with expectations.
	''' Raises TestComplete after tests.
	''' Throws <see cref="Assert.Fail(string)"/> if error or cancelled
	''' using <see cref="queryDescription"/> in the message.
	''' </summary>
	<System.Runtime.CompilerServices.Extension> _
	Public Function TestTheResults(ByVal op As EntityQueryCoroutineOperation, ByVal action As Action(Of IEnumerable), Optional ByVal queryDescription As String = Nothing) As EntityQueryCoroutineOperation

	  If String.IsNullOrWhiteSpace(queryDescription) Then
		queryDescription = op.TargetOperationName
	  End If

	  AddHandler op.Completed, Sub(s, args)
		ReportFailures(args, queryDescription)
		If Nothing IsNot action Then
			action(args.Results)
		End If
		WaitFor.TestComplete()
	  End Sub

	  Return op
	End Function

	''' <summary>
	''' Test the result of a successful <see cref="EntityQuery{T}"/>
	''' from a <see cref="EntityQueryCoroutineOperation{T}"/>
	''' for conformance with expectations.
	''' Raises TestComplete after tests.
	''' Throws <see cref="Assert.Fail(string)"/> if error or cancelled
	''' using <see cref="queryDescription"/> in the message.
	''' </summary>
	<System.Runtime.CompilerServices.Extension> _
	Public Function TestTheResults(Of T)(ByVal op As EntityQueryCoroutineOperation(Of T), ByVal action As Action(Of IEnumerable(Of T)), Optional ByVal queryDescription As String = Nothing) As EntityQueryCoroutineOperation(Of T)

	  If String.IsNullOrWhiteSpace(queryDescription) Then
		queryDescription = op.TargetOperationName
	  End If

	  AddHandler op.Completed, Sub(s, args)
		ReportFailures(args, queryDescription)
		If Nothing IsNot action Then
			action(args.Results)
		End If
		WaitFor.TestComplete()
	  End Sub

	  Return op
	End Function

	''' <summary>
	''' Test the result of a successful <see cref="EntityScalarQuery{T}"/>
	''' from a <see cref="EntityScalarQueryCoroutineOperation{T}"/>
	''' for conformance with expectations.
	''' Raises TestComplete after tests.
	''' Throws <see cref="Assert.Fail(string)"/> if error or cancelled
	''' using <see cref="queryDescription"/> in the message.
	''' </summary>
	<System.Runtime.CompilerServices.Extension> _
	Public Function TestTheResult(Of T)(ByVal op As EntityScalarQueryCoroutineOperation(Of T), ByVal action As Action(Of T), Optional ByVal queryDescription As String = Nothing) As EntityScalarQueryCoroutineOperation(Of T)

	  If String.IsNullOrWhiteSpace(queryDescription) Then
		queryDescription = op.TargetOperationName
	  End If

	  AddHandler op.Completed, Sub(s, args)
		ReportFailures(args, queryDescription)
		If Nothing IsNot action Then
			action(args.Result)
		End If
		WaitFor.TestComplete()
	  End Sub

	  Return op
	End Function

	Private Sub ReportFailures(Of T As INotifyCompletedArgs)(ByVal args As T, ByVal operationDescription As String)
	  operationDescription = If(operationDescription, "Operation")
	  If args.Cancelled Then
		Assert.Fail("Unexpected cancellation of " & operationDescription)
	  End If
	  If Nothing IsNot args.Error Then
		args.IsErrorHandled = True ' we're handling by asserting failure
        Assert.Fail(operationDescription & " failed with exception: " & args.Error.ToString())
	  End If
	End Sub

	''' <summary>Expect a particular exception. Test for it.</summary>
	''' <typeparam name="TException">Type of expected exception</typeparam>
	''' <param name="op">The DevForce async <see cref="BaseOperation"/> object</param>
	''' <param name="operationDescription">
	''' Description of the operation being performed (optional).</param>
	''' <param name="errorMessageContains">
	''' Text fragment that must appear in the exception message (optiona).
	''' </param>
	''' <returns></returns>
	<System.Runtime.CompilerServices.Extension> _
	Public Function ExpectException(Of TException As Exception)(ByVal op As BaseOperation, Optional ByVal operationDescription As String = Nothing, Optional ByVal errorMessageContains As String = Nothing) As BaseOperation

	  AddHandler op.Completed, Sub(sender, args)

		If String.IsNullOrWhiteSpace(operationDescription) Then
		  operationDescription = op.GetType().Name
		End If

		Dim failMsg = String.Format("Expected ""{0}"" exception from {1} ", GetType(TException).Name, operationDescription)

		If args.HasError Then
		  args.CheckExpectedException(Of TException)(failMsg, operationDescription, errorMessageContains)
		ElseIf args.Cancelled Then
		  Assert.Fail(failMsg & "but operation was cancelled instead.")
		Else
		  Assert.Fail(failMsg & "but there was no exception at all.")
		End If
	  End Sub

	  Return op
	End Function

	''' <summary>Expect a particular exception. Test for it.</summary>
	''' <typeparam name="TException">Type of expected exception</typeparam>
	''' <param name="op">The DevForce async <see cref="BaseOperation"/> object</param>
	''' <param name="operationDescription">
	''' Description of the operation being performed (optional).</param>
	''' <param name="errorMessageContains">
	''' Text fragment that must appear in the exception message (optiona).
	''' </param>
	''' <returns></returns>
	<System.Runtime.CompilerServices.Extension> _
	Public Function ExpectException(Of TException As Exception)(ByVal op As CoroutineOperationWrapper, Optional ByVal operationDescription As String = Nothing, Optional ByVal errorMessageContains As String = Nothing) As CoroutineOperationWrapper

	  op.WhenCompleted(Sub(args)
		  If String.IsNullOrWhiteSpace(operationDescription) Then
			  operationDescription = op.TargetOperationName
		  End If
		  Dim failMsg = String.Format("Expected ""{0}"" exception from {1} ", GetType(TException).Name, operationDescription)
		  If args.Cancelled Then
			  Assert.Fail(failMsg & "but operation was cancelled instead.")
		  ElseIf Nothing IsNot args.Error Then
			  args.CheckExpectedException(Of TException)(failMsg, operationDescription, errorMessageContains)
		  Else
			  Assert.Fail(failMsg & "but there was no exception at all.")
		  End If
	  End Sub)

	  Return op
	End Function


	<System.Runtime.CompilerServices.Extension> _
	Private Sub CheckExpectedException(Of TException)(ByVal args As INotifyCompletedArgs, ByVal failMsg As String, ByVal operationDescription As String, ByVal errorMessageContains As String)

	  Dim err = args.Error
      Dim errDescription = err.GetType().Name & " - " & err.Message
	  If TypeOf err Is TException Then
		' got expected exception; now does it contain the expected error message fragment?
		If (Not String.IsNullOrEmpty(errorMessageContains)) AndAlso (Not err.Message.Contains(errorMessageContains)) Then

		  failMsg = String.Format("{0}with message containing ""{1}"" but actual exception was ""{2}"".", failMsg, errorMessageContains, errDescription)

		  Assert.Fail(failMsg)
		Else
		  Console.WriteLine(operationDescription & " threw expected exception: " & errDescription)
		  args.IsErrorHandled = True
		  WaitFor.TestComplete()
		End If
	  Else
		Assert.Fail(failMsg & "but actual exception was """ & errDescription & """.")
	  End If
	End Sub
  End Module

End Namespace