Imports System.Reflection
Imports IdeaBlade.EntityModel

Namespace IdeaBlade.TestFramework

  ''' <summary>
  ''' Provides access to the SilverlightTest Enqueue methods on the current thread.
  ''' </summary>
  Public Module WaitFor

	''' <summary>
	''' Add a Callback method into the test task queue. 
	''' Similar to the PumpMessages(...) call, with the difference being 
	''' that there is no longer a single requirement: 
	''' you can enqueue several callback methods and other test tasks, 
	''' all of which will execute before the test completes and/or the engine continues.
	''' </summary>
	''' <param name="testCallback">
	''' Void-returning delegate, anonymous delegates work fine too.
	''' </param>
	Public Sub Callback(ByVal testCallback As Action)
	  DevForceTest.TestClassBase.EnqueueCallback(testCallback)
	End Sub


	Public Sub Delay(ByVal milliseconds As Double)
	  DevForceTest.TestClassBase.EnqueueDelay(milliseconds)
	End Sub
	''' <summary>
	''' Delay a minimum amount of time before continuing. Similar to a sleep call that is not a blocking call.
	''' </summary>
	''' <param name="delay">
	''' The minimum time span to wait before continuing
	''' </param>
	Public Sub Delay(ByVal delay As TimeSpan)
	  DevForceTest.TestClassBase.EnqueueDelay(delay)
	End Sub
	' Based on Justin Angel: http://blogs.silverlight.net/blogs/justinangel/archive/2009/02/25/silverlight-unit-testing-rhinomocks-unity-and-resharper.aspx
	<System.Runtime.CompilerServices.Extension> _
	Public Function [Event](Of T)(ByVal objectToWaitForItsEvent As T, ByVal eventName As String) As T

	  Dim eventInfo As EventInfo = objectToWaitForItsEvent.GetType().GetEvent(eventName)

	  Dim eventRaised As Boolean = False

	  If GetType(RoutedEventHandler).IsAssignableFrom(eventInfo.EventHandlerType) Then
		eventInfo.AddEventHandler(objectToWaitForItsEvent, CType(Sub() eventRaised = True, RoutedEventHandler))

	  ElseIf GetType(EventHandler).IsAssignableFrom(eventInfo.EventHandlerType) Then
		eventInfo.AddEventHandler(objectToWaitForItsEvent, CType(Sub() eventRaised = True, EventHandler))
	  End If

	  DevForceTest.TestClassBase.EnqueueConditional(Function() eventRaised)

	  Return objectToWaitForItsEvent
	End Function

	''' <summary>
	''' Wait until the DevForce <see cref="BaseOperation.IsCompleted"/> flag is true.
	''' Uses the <see cref="Conditional"/> method.
	''' </summary>
	''' <param name="op">Operation object returned from a DevForce async method.</param>
	Public Sub IsCompleted(ByVal op As BaseOperation)
	  DevForceTest.TestClassBase.EnqueueConditional(Function() op.IsCompleted OrElse op.HasError OrElse op.Cancelled)
	End Sub
	''' <summary>
	''' Instructs the test task queue to wait until the conditional call returns True 
	''' to continue executing other test tasks and/or ending the test method.
	''' </summary>
	''' <param name="conditional">
	''' Conditional method or delegate. Test will halt until this condition returns True.
	''' </param>
	Public Sub Conditional(ByVal conditional As Func(Of Boolean))
	  DevForceTest.TestClassBase.EnqueueConditional(conditional)
	End Sub
	''' <summary>
	''' Enqueue a test task which calls the <see cref="TestCompleteRightNow"/> method.
	''' </summary>
	Public Sub TestComplete()
	  DevForceTest.TestClassBase.EnqueueTestComplete()
	End Sub
	''' <summary>
	''' Signal that a test is complete when using Async testing. 
	''' WARNING: If you use your own methods for completing, 
	''' such as an HtmlTimer or other threading method, 
	''' it is possible that this call will occur *AFTER* the test has timed out 
	''' when using Timeouts. 
	''' As such, be very careful as you could complete the call to the *next* test.
	''' </summary>
	Public Sub TestCompleteRightNow()
	  DevForceTest.TestClassBase.TestComplete()
	End Sub
  End Module
End Namespace