﻿Imports IdeaBlade.EntityModel
Imports Microsoft.Silverlight.Testing
Imports Microsoft.VisualStudio.TestTools.UnitTesting
Imports SimpleNorthwindModel

' ReSharper disable InconsistentNaming
' ReSharper disable RedundantDefaultFieldInitializer
' ReSharper disable RedundantArgumentName

Namespace WorkingWithEntities.EntityStateChangeTests

	<TestClass>
	Public Class When_toying_with_mix_of_customers
		Inherits EntityStateChangeTestBase
		<TestInitialize>
		Public Sub TestInitialize()
			manager = New EntityManager(shouldConnect:= False)
			MakeTestCustomers()
		End Sub

		<TestMethod, Tag("EntityChange")>
		Public Sub then_manager_EntityChanging_reports_changes_to_entity()
			Dim acts = New List(Of EntityAction)()
			Dim chgCnt As Integer = 0
			AddHandler manager.EntityChanging, Sub(s, e) acts.Add(e.Action)
			Dim cust = manager.GetQuery(Of Customer)().First()
			chgCnt += 1
			cust.CompanyName = "Sally"
			chgCnt += 1
			cust.CompanyName = "Robin" ' second mod is counted as well
			chgCnt += 1
			cust.EntityAspect.RejectChanges()
			chgCnt += 1
			manager.RemoveEntity(cust)
			chgCnt += 1
			manager.AttachEntity(cust)
			chgCnt += 1
			cust.EntityAspect.Delete()
			chgCnt += 1
			cust.EntityAspect.AcceptChanges()
			chgCnt += 1

			Assert.AreEqual(chgCnt,acts.Count, "changes detected should equal # of changes")
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub testCustomers_are_in_expected_states()
			AssertIsAdded(addedCustomer)
			AssertIsUnchanged(unchangedCustomer)
			AssertIsModified(modifiedCustomer)
			AssertIsDeleted(deletedCustomer)
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub change_to_unchanged_entity_isModified()
			unchangedCustomer.CompanyName = "abracadabra"
			AssertIsModified(unchangedCustomer)
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub then_delete_of_added_entity_isDetached()
			addedCustomer.EntityAspect.Delete()
			AssertIsDetached(addedCustomer)
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub undo_of_added_entity_isDetached()
			addedCustomer.EntityAspect.RejectChanges()
			AssertIsDetached(addedCustomer)
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub isChanged_of_detached_unchanged_entity_is_true()
			' IsChanged and HasChanges() are defined in terms of EntityState
			' An entity IsChanged if it's EntityState is other than Unchanged.
			unchangedCustomer.EntityAspect.RemoveFromManager()
			Assert.IsTrue(addedCustomer.EntityAspect.IsChanged)
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub delete_of_existing_entity_isDeleted()
			unchangedCustomer.EntityAspect.Delete()
			AssertIsDeleted(unchangedCustomer, "unchanged should be deleted")

			modifiedCustomer.EntityAspect.Delete()
			AssertIsDeleted(modifiedCustomer, "modified should be deleted")
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub undo_of_modified_entity_isUnchanged()
			modifiedCustomer.EntityAspect.RejectChanges()
			AssertIsUnchanged(modifiedCustomer)
		End Sub

	End Class

	<TestClass>
	Public Class When_setting_navigation_properties
		Inherits EntityStateChangeTestBase
		<TestInitialize>
		Public Sub TestInitialize()
			manager = New EntityManager(shouldConnect:= False)
			unchangedCustomer = MakeUnchangedCustomer("Test Company")
			newOrder = New Order()
		End Sub

		Private newOrder As Order

		<TestMethod, Tag("EntityState")>
		Public Sub assigning_ref_nav_adds_assignee_to_manager()
			AssertIsDetached(newOrder, "newOrder should be detached at start.")

			newOrder.Customer = unchangedCustomer

			AssertIsAdded(newOrder, "newOrder should be added after nav assignment.")

			Assert.AreSame(newOrder.EntityAspect.EntityManager, unchangedCustomer.EntityAspect.EntityManager, "EntityManagers should be the same")

			AssertIsUnchanged(unchangedCustomer, "adding to unchangedCustomer's orders shouldn't modify the customer itself")
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub assigning_ref_nav_id_wont_add_assignee_to_manager()
			' Lesson: if you DO NOT want the new entity to be added
			' (e.g., you are making a test or design entity) 
			' then DO NOT set one of its reference navigation properties.
			' Set the corresponding FK id property instead.

			newOrder.CustomerID = unchangedCustomer.CustomerID

			AssertIsDetached(newOrder, "newOrder should be detached after FK assignment.")

			manager.AttachEntity(newOrder)

			AssertIsUnchanged(newOrder, "newOrder should be unchanged after attach.")

			Dim orders = unchangedCustomer.Orders
			Assert.IsTrue(orders.Contains(newOrder), "newOrder should belong to unchangedCustomer.")
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub assigning_ref_nav_of_detached_cust_wont_add_order_to_manager()
			Dim newCust = New Customer With {.CustomerID = Guid.NewGuid()}

			newOrder.Customer = newCust ' NEITHER has an EntityManager!

			AssertIsDetached(newCust, "newCust should be detached.")
			AssertIsDetached(newOrder, "newOrder should be detached.")

			manager.AttachEntity(newCust) ' This works because GUID Id
			manager.AttachEntity(newOrder) ' Danger: Id always 0 because auto-inc Id !!

			AssertIsUnchanged(newCust, "newCust should be unchanged after attach.")
			AssertIsUnchanged(newOrder, "newOrder should be unchanged after attach.")

			Dim orders = newCust.Orders
			Assert.IsTrue(orders.Contains(newOrder), "newOrder should belong to unchangedCustomer.")
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub adding_detached_entity_to_existing_entity_collection_adds_to_manager()

			unchangedCustomer.Orders.Add(newOrder)

			AssertIsAdded(newOrder, "newOrder should be added after nav assignment.")

			Assert.AreSame(newOrder.EntityAspect.EntityManager, unchangedCustomer.EntityAspect.EntityManager, "EntityManagers should be the same")

			AssertIsUnchanged(unchangedCustomer, "adding to unchangedCustomer's orders shouldn't modify the customer itself")
		End Sub
	End Class

	<TestClass>
	Public Class When_using_EntityState_Set_methods
		Inherits EntityStateChangeTestBase
		<TestInitialize>
		Public Sub TestInitialize()
			manager = New EntityManager(shouldConnect:= False)
			MakeTestCustomers()
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub unchanged_becomes_Added_when_call_SetAdded()
			unchangedCustomer.EntityAspect.SetAdded()
			AssertIsAdded(unchangedCustomer)
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub unchanged_becomes_Modified_when_call_SetModified()
			unchangedCustomer.EntityAspect.SetModified()
			AssertIsModified(unchangedCustomer)
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub can_chain_SetAdded_SetModified_Undo()
			unchangedCustomer.EntityAspect.SetAdded()
			AssertIsAdded(unchangedCustomer, " after SetAdded")
			unchangedCustomer.EntityAspect.SetModified()
			AssertIsModified(unchangedCustomer, " after SetModified")
			unchangedCustomer.EntityAspect.RejectChanges()
			AssertIsUnchanged(unchangedCustomer, " after Undo")
		End Sub
	End Class

	<TestClass>
	Public Class When_looking_in_entity_cache
		Inherits EntityStateChangeTestBase
		<TestInitialize>
		Public Sub TestInitialize()
			manager = New EntityManager(shouldConnect:= False)
			MakeTestCustomers()
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub can_find_all_test_customers()
			Dim entities = manager.FindEntities(EntityState.AllButDetached).OfType(Of Entity)()

			Assert.IsTrue(entities.Contains(addedCustomer), "missing added")
			Assert.IsTrue(entities.Contains(unchangedCustomer), "missing unchanged")
			Assert.IsTrue(entities.Contains(modifiedCustomer), "missing modified")
			Assert.IsTrue(entities.Contains(deletedCustomer), "missing deleted")
			Assert.AreEqual(4, entities.Count(), "Should have exactly 4 in cache")
		End Sub


		<TestMethod, Tag("EntityState")>
		Public Sub can_find_all_changed_test_customers()
			Dim entities = manager.FindEntities(EntityState.AnyAddedModifiedOrDeleted).OfType(Of Entity)()

			Assert.IsFalse(entities.Contains(unchangedCustomer), "found unchanged")

			Assert.IsTrue(entities.Contains(addedCustomer), "missing added")
			Assert.IsTrue(entities.Contains(modifiedCustomer), "missing modified")
			Assert.IsTrue(entities.Contains(deletedCustomer), "missing deleted")
			Assert.AreEqual(3, entities.Count(), "Should have exactly 3 changed in cache")
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub dont_find_detached_in_cache()
			unchangedCustomer.EntityAspect.RemoveFromManager()

			Dim entities = manager.FindEntities(EntityState.AnyAddedModifiedOrDeleted).OfType(Of Entity)()

			Assert.IsFalse(entities.Contains(unchangedCustomer), "found unchanged")
		End Sub
	End Class

	<TestClass>
	Public Class When_combining_EntityState_enums
		Inherits EntityStateChangeTestBase
		<TestInitialize>
		Public Sub TestInitialize()
			manager = New EntityManager(shouldConnect:= False)
			MakeTestCustomers()
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub can_combine_EntityState_enums_to_check_entity_state()
			Dim currentState = unchangedCustomer.EntityAspect.EntityState
			If (currentState And (EntityState.Added Or EntityState.Modified)) > 0 Then
				Assert.Fail("Shouldn't be added or modified")
			End If
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub can_use_EntityState_extensions_to_check_entity_state()
			Dim currentState = unchangedCustomer.EntityAspect.EntityState
			If currentState.IsAddedOrModified() Then
				Assert.Fail("Shouldn't be added or modified")
			End If
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub can_combine_EntityState_enums_to_find_changed_entities_in_Cache()
			Const desiredStates As EntityState = EntityState.Added Or EntityState.Modified Or EntityState.Deleted
			Dim changedEntities = manager.FindEntities(Of Customer)(desiredStates)
			Assert.AreEqual(3, changedEntities.Count())
		End Sub

		<TestMethod, Tag("EntityState")>
		Public Sub can_use_predefined_combo_enum_to_find_changed_entities_in_Cache()
			Dim changedEntities = manager.FindEntities(Of Customer)(EntityState.AnyAddedModifiedOrDeleted)
			Assert.AreEqual(3, changedEntities.Count())
		End Sub
	End Class

	Public MustInherit Class EntityStateChangeTestBase

		Protected manager As EntityManager

		#Region "EntityStateAsserts"

		Protected Sub AssertIsAdded(ByVal entity As IEntity, Optional ByVal message As String = Nothing)
			Assert.AreEqual(EntityState.Added, entity.EntityAspect.EntityState, message)
		End Sub
		Protected Sub AssertIsDeleted(ByVal entity As IEntity, Optional ByVal message As String = Nothing)
			Assert.AreEqual(EntityState.Deleted, entity.EntityAspect.EntityState, message)
		End Sub
		Protected Sub AssertIsDetached(ByVal entity As IEntity, Optional ByVal message As String =Nothing)
			Assert.AreEqual(EntityState.Detached, entity.EntityAspect.EntityState, message)
		End Sub
		Protected Sub AssertIsModified(ByVal entity As IEntity, Optional ByVal message As String = Nothing)
			Assert.AreEqual(EntityState.Modified, entity.EntityAspect.EntityState, message)
		End Sub
		Protected Sub AssertIsUnchanged(ByVal entity As IEntity, Optional ByVal message As String = Nothing)
			Assert.AreEqual(EntityState.Unchanged, entity.EntityAspect.EntityState, message)
		End Sub

		#End Region

		#Region "MakeTestCustomers"

		Protected Sub MakeTestCustomers()
			addedCustomer = MakeAddedCustomer("Added Company")
			unchangedCustomer = MakeUnchangedCustomer("Unchanged Company")
			modifiedCustomer = MakeModifiedCustomer("Modified Company")
			deletedCustomer = MakeDeletedCustomer("Deleted Company")
		End Sub

		Protected addedCustomer As Customer
		Protected unchangedCustomer As Customer
		Protected modifiedCustomer As Customer
		Protected deletedCustomer As Customer

		Protected Function MakeAddedCustomer(ByVal companyName As String) As Customer
			Dim cust = New Customer With {.CompanyName = companyName, .Country = "Freedonia"}
			manager.AddEntity(cust)
			Return cust
		End Function

		Protected Function MakeUnchangedCustomer(ByVal companyName As String) As Customer
			Dim cust = MakeAddedCustomer(companyName)
			cust.EntityAspect.AcceptChanges()
			Return cust
		End Function

		Protected Function MakeModifiedCustomer(ByVal companyName As String) As Customer
			Dim cust = MakeUnchangedCustomer(companyName)
			cust.EntityAspect.SetModified()
			Return cust
		End Function

		Protected Function MakeDeletedCustomer(ByVal companyName As String) As Customer
			Dim cust = MakeUnchangedCustomer(companyName)
			cust.EntityAspect.Delete()
			Return cust
		End Function

		#End Region

	End Class
End Namespace